/* opserv.c - IRC Operator assistant service
 * Copyright 2000-2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: opserv.c,v 1.187.2.88 2002/02/22 03:08:41 entrope Exp $
 */

#include "config.h"
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_SYS_TIMES_H
#include <sys/times.h>
#endif

#include "common.h"
#include "dict.h"
#include "conf.h"
#include "gline.h"
#include "hash.h"
#include "heap.h"
#include "helpfile.h"
#include "log.h"
#include "parse.h"
#include "chanserv.h"
#include "nickserv.h"
#include "global.h"
#include "opserv.h"
#include "policer.h"
#include "recdb.h"
#include "timeq.h"
#include "alias.h"
#include "sockcheck.h"

#define OPSERV_CONF_NAME "services/opserv"
#define OPSERV_DB_NAME "opserv.db"
#define TEMP_DB_NAME "opserv.db.new"
#define OPSERV_HELPFILE_NAME "opserv.help"

#define KEY_ALERT_CHANNEL "alert_channel"
#define KEY_DEBUG_CHANNEL "debug_channel"
#define KEY_UNTRUSTED_MAX "untrusted_max"
#define KEY_PURGE_LOCK_DELAY "purge_lock_delay"
#define KEY_NICK "nick"
#define KEY_TRIGGER "trigger"
#define KEY_JOIN_POLICER "join_policer"
#define KEY_NEW_USER_POLICER "new_user_policer"
#define KEY_COMMANDS "commands"
#define KEY_MIN_LEVEL "min_level"
#define KEY_DISABLED "disabled"
#define KEY_WEIGHT "weight"
#define KEY_GLINES "glines"
#define KEY_REASON "reason"
#define KEY_RESERVES "reserves"
#define KEY_USERNAME "username"
#define KEY_HOSTNAME "hostname"
#define KEY_DESC "description"
#define KEY_PERSISTENT "persistent"
#define KEY_BAD_WORDS "bad"
#define KEY_SECRET_WORDS "secret"
#define KEY_TRUSTED_HOSTS "trusted"
#define KEY_DB_BACKUP_FREQ "db_backup_freq"
#define KEY_OWNER "owner"
#define KEY_GAGS "gags"
#define KEY_ALERTS "alerts"
#define KEY_REACTION "reaction"
#define KEY_DISCRIM "discrim"
#define KEY_WARN "chanwarn"
#define KEY_MAX "max"
#define KEY_TIME "time"
#define KEY_MAX_CLIENTS "max_clients"
#define KEY_LIMIT "limit"
#define KEY_EXPIRES "expires"
#define KEY_STAFF_AUTH_CHANNEL "staff_auth_channel"

#define OSMSG_USER_ACCESS_IS    "$b%s$b (handle $b%s$b) has access level %d."
#define OSMSG_HANDLE_ACCESS_IS  "$b%s$b has access level %d."
#define OSMSG_LEVEL_TOO_LOW	"Your access level is too low to do that."
#define OSMSG_INTERNAL_COMMAND  "$b%s$b appears to be an internal $O command, sorry."
#define OSMSG_DUMBASS		"You may not disable the $b%s$b command."
#define OSMSG_NEED_CHANNEL	"You must specify a channel for `%s'."
#define OSMSG_LEVEL_SET		"`%s' now has access level %d."
#define OSMSG_INVALID_IRCMASK	"`%s' is an invalid IRC hostmask."
#define OSMSG_BAD_HOSTMASK      "`%s' isn't parseable as an IP-based mask."
#define OSMSG_STRING_TOO_LONG	"Some part of your command was too long."
#define OSMSG_ADDED_BAN		"I have banned `%s' from `%s'."
#define OSMSG_GLINE_REQUESTED	"G-line requested by %s."
#define OSMSG_GLINE_ISSUED	"G-line issued for `%s'."
#define OSMSG_GLINE_REMOVED	"G-line removed for `%s'."
#define OSMSG_GLINES_REFRESHED	"All G-lines refreshed."
#define OSMSG_KILL_REQUESTED	"Kill requested by %s."
#define OSMSG_GAG_REQUESTED	"Gag requested by %s."
#define OSMSG_CLEARBANS_DONE	"Cleared all bans from channel `%s'."
#define OSMSG_CLEARMODES_DONE	"Cleared ALL modes from channel `%s'."
#define OSMSG_NO_CHANNEL_MODES	"Channel `%s' had no modes to clear."
#define OSMSG_DEOP_DONE		"Deop'ed the requested lusers."
#define OSMSG_DEOPALL_DONE	"Deop'ed everyone on `%s'."
#define OSMSG_ISON_PREFIX	"`%s' is on:"
#define OSMSG_NO_DEBUG_CHANNEL	"No debug channel has been configured."
#define OSMSG_INVITE_DONE	"Invited `%s' to `%s'."
#define OSMSG_JOIN_DONE		"I have joined `%s'."
#define OSMSG_NOT_ON_CHANNEL	"`%s' does not seem to be on `%s'."
#define OSMSG_KICK_REQUESTED	"Kick requested by %s."
#define OSMSG_KICKALL_DONE	"I have cleared out %s."
#define OSMSG_LEAVE_DONE	"I have left `%s'."
#define OSMSG_MODE_SET		"I have set the modes for `%s'."
#define OSMSG_ALREADY_BANNED	"But ban for `%s' already applies to `%s'."
#define OSMSG_OP_DONE		"I have op'ed people on `%s'."
#define OSMSG_OPERMSG_FORMAT	"Oper notice <%s>: %s"
#define OSMSG_WHOIS_IDENT	"%s (%s@%s) from %d.%d.%d.%d"
#define OSMSG_WHOIS_NICK	"Nick    : %s"
#define OSMSG_WHOIS_HOST	"Host    : %s@%s"
#define OSMSG_WHOIS_IP  	"Real IP : %ld.%ld.%ld.%ld"
#define OSMSG_WHOIS_INFO	"Info    : %s"
#define OSMSG_WHOIS_NUMERIC	"Numeric : %s"
#define OSMSG_WHOIS_SERVER	"Server  : %s"
#define OSMSG_WHOIS_CHANNELS	"Channels:"
#define OSMSG_WHOIS_HIDECHANS   "Channel list omitted for your sanity."
#define OSMSG_OTHER_NOT_AUTHED	"`%s' is not authenticated."
#define OSMSG_READHELP_SUCCESS	"Read help database in "FMT_TIME_T".%03lu seconds."
#define OSMSG_UNBAN_DONE	"Ban(s) removed from channel %s."
#define OSMSG_ALREADY_MODERATED "Channel `%s' is already moderated."
#define OSMSG_CHANNEL_MODERATED	"Channel `%s' is now moderated."
#define OSMSG_ALREADY_UNMODERATED "Channel `%s' is not currently moderated."
#define OSMSG_CHANNEL_UNMODERATED "Channel `%s' is no longer moderated."
#define OSMSG_COMMAND_ACCESS_IS "Minimum access to use $b%s$b is %d."
#define OSMSG_COMMAND_ENABLED	"Command `%s' is now enabled."
#define OSMSG_ADDED_BAD		"Added bad words to the bad-word list."
#define OSMSG_REMOVED_BAD	"Removed bad words from the bad-word list."
#define OSMSG_ADDED_SECRET	"Added secret words to the secret-word list."
#define OSMSG_REMOVED_SECRET	"Removed secret words from the secret-word list."
#define OSMSG_ALREADY_TRUSTED   "Host `%s' is already trusted (use `deltrust' and then `addtrust' to adjust)."
#define OSMSG_ADDED_TRUSTED     "Added trusted hosts to the trusted-hosts list."
#define OSMSG_REMOVED_TRUSTED   "Removed trusted hosts from the trusted-hosts list."
#define OSMSG_CLONE_EXISTS	"Nick `%s' is already in use."
#define OSMSG_NOT_A_HOSTMASK	"The hostmask must be in user@host form."
#define OSMSG_DISABLED_LIST	"Disabled commands: %s"
#define OSMSG_BADWORD_LIST	"Bad words: %s"
#define OSMSG_GLINE_COUNT       "There are %d glines active on the network."
#define OSMSG_BEGIN_LINKS	"Server link info:"
#define OSMSG_LINK_INFO		"Hub: %s Link: %s"
#define OSMSG_END_LINKS		"End of link info."
#define OSMSG_MAX_CLIENTS	"Max clients: %d at %s"
#define OSMSG_NETWORK_INFO	"Total users: %d (%d invisible, %d opers)"
#define OSMSG_NETWORK_SERVER	"%-27s Users: %d (%3.3g%%)"
#define OSMSG_NETWORK_SERVER_2  "%-27s Numeric %s: %ld Linked: %s"
#define OSMSG_STATS_OPER_LIST	"Current list of operators online:"
#define OSMSG_RESERVED_LIST	"List of reserved nicks:"
#define OSMSG_RESERVED_NICK_ITEM "%s"
#define OSMSG_SECRET_LIST	"List of secret channel words:"
#define OSMSG_WORD_IS_SECRET	"%s"
#define OSMSG_TRUSTED_LIST	"List of trusted hosts:"
#define OSMSG_HOST_IS_TRUSTED	"%s (limit %lu, expires %s: %s)"
#define OSMSG_UPTIME_STATS	"Uptime: %s (%u lines processed, CPU time %.2fu/%.2fs)"
#define OSMSG_NO_LAST_WRITE     "Last Database Write: Never"
#define OSMSG_LAST_WRITE        "Last Database Write: %s Ago"
#define OSMSG_SOCKCHECK_STATS	"Since booting, I have checked %d clients for illicit proxies, and detected %d proxy hosts.\nI am currently checking %d clients (out of %d max) and have a backlog of %d more to start on.\nI currently have %d hosts cached.\nI know how to detect %d kinds of proxies."
#define OSMSG_UNKNOWN_STATS	"Unknown statistics type `%s'"
#define OSMSG_ADDRESS_QUEUED	"`%s' is now queued to be proxy-checked."
#define OSMSG_HOST_CLEARED	"'%s' was cleared from the cached hosts list."
#define OSMSG_HOST_NOTCLEARED	"'%s' was not cached and therefore wasn't cleared."
#define OSMSG_HOST_BUSY		"'%s' is currently being checked, unable to clear."
#define OSMSG_NO_SUCH_ADDRESS	"Unable to resolve `%s' to an IP address."
#define OSMSG_LINE_DUMPED	"Raw line sent."
#define OSMSG_RAW_PARSE_ERROR	"Error parsing raw line (not dumping to uplink)"
#define OSMSG_COLLIDED_NICK	"Now temporarily holding nick `%s'"
#define OSMSG_RESERVED_NICK	"Now reserving nick `%s'"
#define OSMSG_NICK_RESERVED	"Nick is reserved."
#define OSMSG_NICK_UNRESERVED	"No longer reserving nick `%s'."
#define OSMSG_NOT_RESERVED	"Nick `%s' is not reserved."
#define OSMSG_ILLEGAL_REASON    "This channel is illegal."
#define OSMSG_ILLEGAL_CHANNEL	"$b%s$b is an ILLEGAL channel. Do not re-join it."
#define OSMSG_JOIN_FLOOD	"Warning: Possible join flood in %s (currently %d users)."
#define OSMSG_NEW_USER_FLOOD    "Warning: Possible new-user flood."
#define OSMSG_CLONE_WARNING     "WARNING: You have connected the maximum permitted number of clients from one IP address (clones).  If you connect any more, your host will be banned from the network."
#define OSMSG_TOO_MANY_CLONES   "Excessive connections from a single host."
#define OSMSG_CLONE_ADDED	"Added clone `%s'."
#define OSMSG_NOT_A_CLONE	"Har har.  `%s' isn't a clone."
#define OSMSG_PART_REASON	"%s has no reason."
#define OSMSG_CLONE_REMOVED	"Removed clone `%s'."
#define OSMSG_CLONE_JOINED	"`%s' has joined `%s'."
#define OSMSG_CLONE_PARTED	"`%s' has left `%s'."
#define OSMSG_OPS_GIVEN		"I have given ops on `%s' to `%s'."
#define OSMSG_CLONE_SAID	"`%s' has spoken to `%s'."
#define OSMSG_UNKNOWN_SUBCOMMAND "`%s' is not a valid subcommand of `%s'."
#define OSMSG_WRITE_SUCCESS	"$O db write completed (in "FMT_TIME_T".%03lu seconds)."
#define OSMSG_WRITEALL_SUCCESS	"All db write completed (in "FMT_TIME_T".%03lu seconds)."
#define OSMSG_UNKNOWN_OPTION    "`%s' has not been set."
#define OSMSG_OPTION_IS         "`%s' is set to '%s'."
#define OSMSG_OPTION_ROOT       "The following keys can be queried:"
#define OSMSG_OPTION_LIST       "`%s' contains the following values:"
#define OSMSG_OPTION_KEYS       "`%s' contains the following keys:"
#define OSMSG_OPTION_LIST_ENTRY "`%s'"
#define OSMSG_SET_NOT_SET       "'%s' does not exist, and cannot be set."
#define OSMSG_SET_BAD_TYPE      "'%s' is not a string, and cannot be set."
#define OSMSG_SET_SUCCESS       "'%s' has been set to '%s'."
#define OSMSG_NOT_IP            "You must provide a valid IPv4 address."
#define OSMSG_BAD_ACTION        "Unrecognized trace action '%s'."
#define OSMSG_GSYNC_RUNNING     "Synchronizing glines from %s."
#define OSMSG_GTRACE_FORMAT     "%s (by %s, expires %s): %s"
#define OSMSG_ALIAS_CREATED	"Added %s alias '%s' for '%s'."
#define OSMSG_ALIAS_DELETED	"Deleted %s alias '%s'."
#define OSMSG_ALIAS_MODIFIED	"Modified %s alias '%s' to call '%s'."
#define OSMSG_NO_SUCH_SERVICE	"Could not find service '%s'."
#define OSMSG_NO_SUCH_COMMAND	"Could not find %s command '%s'."
#define OSMSG_NO_SUCH_ALIAS	"Could not find %s alias '%s'."
#define OSMSG_ALIAS_CREATE_FAIL	"Could not create %s alias '%s'."
#define OSMSG_WONT_ALIAS_CMD	"Service %s already has a command named '%s'."
#define OSMSG_SELF_ALIAS	"Aliases must point to a different command."
#define OSMSG_SERVICE_COMMANDS	"$b%s$b Aliases:"
#define OSMSG_COMMAND_ALIASES	"$b%s$b: %s"
#define OSMSG_GAG_APPLIED	"Gagged '%s', affecting %d users."
#define OSMSG_GAG_ADDED		"Gagged '%s'."
#define OSMSG_REDUNDANT_GAG	"Gag '%s' is redundant."
#define OSMSG_GAG_NOT_FOUND    	"Could not find gag '%s'."
#define OSMSG_NO_GAGS    	"No gags have been set."
#define OSMSG_UNGAG_APPLIED	"Ungagged '%s', affecting %d users."
#define OSMSG_UNGAG_ADDED      	"Ungagged '%s'."
#define OSMSG_GAG_HEADER    	"Mask:\tOwner:\tExpires:\tReason:"
#define OSMSG_GAG_DATA    	"%s\t%s\t%s\t%s"
#define OSMSG_TIMEQ_INFO        "%u events in timeq; next in %lu seconds."
#define OSMSG_DEFPROXY_FAILED   "Proxy definition failed: %s"
#define OSMSG_DEFPROXY_SUCCESS  "New proxy type defined."
#define OSMSG_ALERT_TRIGGERED	"Alert '%s' triggered by user '%s' (%s)."
#define OSMSG_ALERT_EXISTS	"An alert named '%s' already exists!"
#define OSMSG_UNKNOWN_REACTION	"Unknown reaction to alert '%s'."
#define OSMSG_ADDED_ALERT	"Added alert named '%s'."
#define OSMSG_REMOVED_ALERT	"Deleted alert named '%s'."
#define OSMSG_NO_SUCH_ALERT	"No alert named '%s' could be found."
#define OSMSG_ALERT_IS		"%s (by %s, reaction %s): %s"
#define OSMSG_ALERTS_LIST	"Current $O alerts:"
#define OSMSG_COMMAND_TIME	"Command `%s' finished in "FMT_TIME_T".%06lu seconds."
#define OSMSG_MISSING_SUPPORT	"This command cannot be used due to lack of support at the operating system level." 
#define OSMSG_REHASH_COMPLETE	"Completed rehash of configuration database."
#define OSMSG_REHASH_FAILED	"Rehash of configuration database failed, previous configuration is intact."
#define OSMSG_REOPEN_COMPLETE	"Closed and reopened all log files."
#define OSMSG_NUMERIC_COLLIDE   "That numeric is already in use."
#define OSMSG_NAME_COLLIDE      "That name is already in use."
#define OSMSG_SRV_CREATE_FAILED "Server creation failed -- check log files."
#define OSMSG_SERVER_JUPED      "Added new jupe server %s."
#define OSMSG_SERVER_NOT_JUPE   "That server is not a juped server."
#define OSMSG_MODLOG_FAILED     "Logging modification failed."
#define OSMSG_MODLOG_PASSED     "Logging modification succeeded."
#define OSMSG_LOG_SEVERITIES    "Logged severities for %s: %s"
#define OSMSG_WARN_ADDED	"Added channel activity warning for '%s' (%s)"
#define OSMSG_WARN_EXISTS	"Channel activity warning for '%s' already exists."
#define OSMSG_WARN_DELETE	"Deleted channel activity warning for '%s'"
#define OSMSG_WARN_NOEXIST	"Channel activity warning for '%s' does not exist."
#define OSMSG_WARN_LISTSTART	"Channel activity warnings:"
#define OSMSG_WARN_LISTEND	"End of activity warning list."
#define OSMSG_INDEX_FORMAT	"%-12s (%d)%c  "
#define OSMSG_CURRENT_UPLINK	"$b%s$b is already the current uplink."
#define OSMSG_INVALID_UPLINK	"$b%s$b is not a valid uplink name."

#define MAX_CHANNELS_WHOIS	50

#define OPSERV_FUNC(NAME) int NAME(struct userNode *user, struct chanNode *channel, unsigned int argc, unsigned char *argv[])
typedef OPSERV_FUNC(opserv_func_t);

static OPSERV_FUNC(cmd_help);

#define OPSERV_SYNTAX() 	opserv_help(user, argv[0], 0)
#define OPSERV_NEED_CHANNEL()	do { if (!channel) { \
    opserv_notice(user, MSG_INVALID_CHANNEL); \
    OPSERV_SYNTAX(); \
    return 0; } } while (0)
#define OPSERV_MIN_PARMS(N, NEED_CHANNEL) do { \
  if (argc < N) { \
    opserv_notice(user, MSG_MISSING_PARAMS, argv[0]); \
    OPSERV_SYNTAX(); \
    return 0; } \
  if (NEED_CHANNEL) OPSERV_NEED_CHANNEL(); } \
  while (0)

typedef int (*discrim_search_func)(struct userNode *match, void *extra);

struct userNode *opserv;
struct helpfile *opserv_helpfile;
dict_t opserv_chan_warn;

static dict_t opserv_func_dict;
static dict_t opserv_reserved_nick_dict;
static struct string_list *opserv_bad_words, *opserv_secret_words;
static dict_t opserv_trusted_hosts;
static dict_t opserv_hostinfo_dict;
static unsigned int new_user_flood;
static dict_t opserv_user_alerts;
static dict_t opserv_nick_based_alerts;
static time_t last_db_write;

extern int quit_services;
extern struct server *servers[64*64];

static struct {
    struct chanNode *debug_channel;
    struct chanNode *alert_channel;
    struct chanNode *staff_auth_channel;
    struct policer_params *join_policer_params;
    struct policer *new_user_policer;
    unsigned long db_backup_frequency;
    unsigned long untrusted_max;
    unsigned long purge_lock_delay;
    unsigned char trigger_char;
} opserv_conf;

struct opserv_cmd {
    unsigned int min_level;
    int disabled;
    opserv_func_t *func;
    double weight;
};

struct trusted_host {
    const char *hostname;
    unsigned long limit;
    time_t expires;
    unsigned char reason[MAXLEN];
};

struct gag_entry {
    char *mask;
    char *owner;
    char *reason;
    time_t expires;
    struct gag_entry *next;
};

static struct gag_entry *gagList;

struct opserv_hostinfo {
    unsigned long client_count;
};

typedef struct opservDiscrim {
    unsigned long limit;
    struct chanNode *channel;
    char *mask_nick, *mask_ident, *mask_host, *mask_info, *server;
    unsigned long ip_addr, ip_mask, min_level, max_level;
    unsigned long match_opers : 1;
    time_t last;
    char *reason;
} *discrim_t;

struct discrim_and_source {
    discrim_t discrim;
    struct userNode *source;
};

static discrim_t opserv_discrim_create(struct userNode *user, unsigned int argc, unsigned char *argv[]);
static unsigned int opserv_discrim_search(discrim_t discrim, discrim_search_func dsf, void *data);
static int gag_helper_func(struct userNode *match, void *extra);

typedef enum {
    REACT_NOTICE,
    REACT_GLINE,
} opserv_alert_reaction;

struct opserv_user_alert {
    char *owner;
    char *text_discrim;
    discrim_t discrim;
    opserv_alert_reaction reaction;
};

/* funny type to make it acceptible to dict_set_free_data, far below */
static void
opserv_free_user_alert(void *data)
{
    struct opserv_user_alert *alert = data;
    free(alert->owner);
    free(alert->text_discrim);
    free(alert->discrim);
    free(alert);
}

const char *OSMSG_VERSION_ID = "$Id: opserv.c,v 1.187.2.88 2002/02/22 03:08:41 entrope Exp $";

#define opserv_notice(target, format...) send_message(target , opserv , ## format)
#define opserv_debug(format...) do { if (opserv_conf.debug_channel) send_channel_notice(opserv_conf.debug_channel , opserv, ## format); } while (0)
#define opserv_alert(format...) do { if (opserv_conf.alert_channel) send_channel_notice(opserv_conf.alert_channel , opserv, ## format); } while (0)

/* This function is called when a channel command (or
 * channel-qualified privmsg) is directed towards OpServ.
 */
static void
opserv_chanmsg(struct userNode* user, struct chanNode *channel, char *text)
{
    struct opserv_cmd *cmd;
    unsigned char *argv[MAXNUMPARAMS], *tmp;
    int argc, argv_shift;

    /* ignore things consistning of empty lines */
    if (!*text) return;
    /* otherwise see if we should listen to them */
    if (!IsOper(user)) {
	opserv_notice(user, MSG_SERVICE_PRIVILEGED, "$O");
	return;
    }
    if (!user->handle_info) {
	opserv_notice(user, MSG_AUTHENTICATE);
	return;
    }
    log(OS_LOG, LOG_INFO, "%s[%s]: %s\n", user->nick, user->handle_info->handle, text);
    if (HANDLE_FLAGGED(user->handle_info, OPER_SUSPENDED)) {
	opserv_notice(user, MSG_OPER_SUSPENDED);
	return;
    }

    argv_shift = 1;
    argc = split_line(text, false, ArrayLength(argv)-argv_shift, argv+argv_shift);
    if(!argc) return;

    if (IsChannelName(argv[argv_shift])) {
        if (argc == 1) return;
        channel = GetChannel(argv[++argv_shift]);
    } else {
        /* first arg isn't a channel name, so it's a command */
        if ((argc > 1) && IsChannelName(argv[argv_shift+1])) {
            channel = GetChannel(argv[argv_shift+1]);
            tmp = argv[argv_shift+1];
            argv[argv_shift+1] = argv[argv_shift];
            argv[argv_shift] = tmp;
            argv_shift++; argc--;
         } else {
            argv[argv_shift-1] = channel ? channel->name : NULL;
         }
    }

    cmd = dict_find(opserv_func_dict, argv[argv_shift], NULL);
    if (!cmd) {
	opserv_notice(user, MSG_COMMAND_UNKNOWN, argv[argv_shift]);
	return;
    }
    if (cmd->min_level > user->handle_info->opserv_level) {
	opserv_notice(user, OSMSG_LEVEL_TOO_LOW);
	return;
    }
    if (cmd->disabled) {
	opserv_notice(user, MSG_COMMAND_DISABLED, argv[argv_shift]);
	return;
    }
    if (cmd->func) {
        cmd->func(user, channel, argc, argv+argv_shift);
    } else {
        opserv_notice(user, OSMSG_INTERNAL_COMMAND, argv[argv_shift]);
    }
}

static void
opserv_privmsg(struct userNode* user, char *command, int server_qualified)
{
    (void)server_qualified;
    opserv_chanmsg(user, 0, command);
}

static void
opserv_help(struct userNode *user, const char *topic, int show_aliases)
{
    char *aliases, *resolved = resolve_alias("OpServ", (char *)topic);

    send_help(user, opserv, opserv_helpfile, resolved);
    if(show_aliases && (aliases = command_aliases("OpServ", resolved)))
    {
	char *c = aliases;

	for(c = aliases; *c; c++)
	{
	    *c = toupper(*c);
	}

	opserv_notice(user, "Aliases: %s", aliases);
	free(aliases);
    }
}

/* A lot of these commands are very similar to what ChanServ can do,
 * but OpServ can do them even on channels that aren't registered.
 */

static OPSERV_FUNC(cmd_access)
{
    struct handle_info *hi;
    char *target;

    (void)channel; (void)argc;
    if (argc > 1) {
	target = argv[1];
    } else {
	target = user->nick;
    }

    hi = smart_get_handle_info(opserv, user, target);
    if (hi) {
        opserv_notice(user, OSMSG_HANDLE_ACCESS_IS, hi->handle, hi->opserv_level);
        return 1;
    }
    return 0;
}

static const char *
oper_numeric(struct chanNode *channel)
{
    struct modeNode *mn;
    return (channel
	    && (mn = GetUserMode(channel, opserv))
	    && (mn->modes & MODE_CHANOP))
	? opserv->numeric : self->numeric;
}

static OPSERV_FUNC(cmd_ban)
{
    char *mask;
    OPSERV_MIN_PARMS(2, true);
    mask = argv[1];
    if (!is_ircmask(mask)) {
	opserv_notice(user, OSMSG_INVALID_IRCMASK, mask);
	return 0;
    }
    AddChannelBan(1, &mask, channel, oper_numeric(channel), time(0), 1);
    opserv_notice(user, OSMSG_ADDED_BAN, mask, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_chaninfo)
{
    char buffer[512];
    struct banNode *ban;
    struct modeNode *moden;
    unsigned int n;

    OPSERV_NEED_CHANNEL(); (void)argc;
    opserv_notice(user, "%s Statistics", channel->name);
    strftime(buffer, sizeof(buffer), "Created on: %a %b %d %H:%M:%S %Y", localtime(&channel->timestamp));
    opserv_notice(user, "%s ("FMT_TIME_T")", buffer, channel->timestamp);
    strcpy(buffer, "Modes: ");
    irc_make_chanmode(channel, buffer+strlen(buffer));
    opserv_notice(user, "%s", buffer);
    if (channel->topic_time > 1) {
        strftime(buffer, sizeof(buffer), "Topic: `%%s$o' set by %%s (%a %b %d %H:%M:%S %Y)", localtime(&channel->topic_time));
        opserv_notice(user, buffer, channel->topic, channel->topic_nick);
    } else {
	irc_fetchtopic(opserv->numeric, channel->name);
	opserv_notice(user, "Topic: (none / not gathered)");
    }
    if (channel->banlist.used) {
	opserv_notice(user, "Bans (%d):", channel->banlist.used);
	for (n = 0; n < channel->banlist.used; n++) {
    	    ban = channel->banlist.list[n];
	    strftime(buffer, sizeof(buffer), "%%s by %%s (%a %b %d %H:%M:%S %Y)", localtime(&ban->set));
	    opserv_notice(user, buffer, ban->ban, ban->who);
	}
    }
    opserv_notice(user, "Users (%d):", channel->members.used);
    for (n=0; n<channel->members.used; n++) {
	moden = channel->members.list[n];
	if (!(moden->modes & MODE_CHANOP)) continue;
	opserv_notice(user, " @%s (%s@%s)", moden->user->nick, moden->user->username, moden->user->hostname);
    }
    for (n=0; n<channel->members.used; n++) {
	moden = channel->members.list[n];
	if ((moden->modes & MODE_CHANOP)
	    || !(moden->modes & MODE_VOICE)) continue;
	opserv_notice(user, " +%s (%s@%s)", moden->user->nick, moden->user->username, moden->user->hostname);
    }
    for (n=0; n<channel->members.used; n++) {
	moden = channel->members.list[n];
	if ((moden->modes & MODE_VOICE)
	    || (moden->modes & MODE_CHANOP)) continue;
	opserv_notice(user, "  %s (%s@%s)", moden->user->nick, moden->user->username, moden->user->hostname);
    }
    return 1;
}

static int
cmd_warnlist_helper(const char *key, void *data, void *extra)
{
    char *reason = data;
    struct userNode *user = extra;
    opserv_notice(user, "%s (%s)", key, reason);
    return 0;
}

static OPSERV_FUNC(cmd_warnlist)
{
    (void)channel; (void)argc; (void)argv;
    opserv_notice(user, OSMSG_WARN_LISTSTART);
    dict_foreach(opserv_chan_warn, cmd_warnlist_helper, user);
    opserv_notice(user, OSMSG_WARN_LISTEND);
    return 1;
}

static OPSERV_FUNC(cmd_warn) 
{
    char *reason, *message;

    /* This was 1 originally...but that doesn't seem right. */
    OPSERV_MIN_PARMS(2, false);

    /* this shouldnt happen */
    if (!opserv_chan_warn) return 0;

    if (argv[-1] && IsChannelName(argv[-1])) {
	char *cmd;
	cmd = dict_find(opserv_chan_warn, argv[-1], NULL);
	if (cmd) {
   	   opserv_notice(user, OSMSG_WARN_EXISTS, argv[-1]);
	   return 0;
        } else {
	   if (argv[2]) {
	       unsplit_string(argv+1, argc-1);
	       reason = strdup(argv[1]);
	   } else reason = strdup("No reason");
	   dict_insert(opserv_chan_warn, strdup(argv[-1]), reason);
	   opserv_notice(user, OSMSG_WARN_ADDED, argv[-1], reason);

	   if (dict_find(channels, argv[-1], NULL)) {
	       message = alloca(strlen(reason) + strlen(argv[-1]) + 55);
	       sprintf(message, "Channel activity warning for channel `%s' (%s).", argv[-1], reason);
	       global_message(MESSAGE_RECIPIENT_OPERS, message);
	   }
	}
    } else {
	opserv_notice(user, OSMSG_NEED_CHANNEL, argv[0]);
	return 0;
    }
    return 1;
}

static OPSERV_FUNC(cmd_unwarn)
{
    (void)argc; (void)channel;

    /* this shouldnt happen */
    if (!opserv_chan_warn) return 0;

    if (argv[-1] && IsChannelName(argv[-1])) {
        char *cmd;
        cmd = dict_find(opserv_chan_warn, argv[-1], NULL);
        if (cmd) {
	   dict_remove(opserv_chan_warn, argv[-1]);
           opserv_notice(user, OSMSG_WARN_DELETE, argv[-1]);
        } else {
           opserv_notice(user, OSMSG_WARN_NOEXIST, argv[-1]);
	   return 0;
        }
    } else {
        opserv_notice(user, OSMSG_NEED_CHANNEL, argv[0]);
	return 0;
    }
    return 1;
}

static OPSERV_FUNC(cmd_clearbans)
{
    char *list[256];
    unsigned int count, n;
    const char *numeric;

    OPSERV_NEED_CHANNEL(); (void)argc;
    numeric = oper_numeric(channel);
    for (count=n=0; n<channel->banlist.used; n++) {
	list[count] = channel->banlist.list[n]->ban;
	if (count == ArrayLength(list)) {
	    DelChannelBan(count, list, channel, numeric, 1);
	    count = 0;
	}
    }
    if (count) {
	DelChannelBan(count, list, channel, numeric, 1);
    }
    opserv_notice(user, OSMSG_CLEARBANS_DONE, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_clearmodes)
{
    char buffer[400];

    OPSERV_NEED_CHANNEL(); (void)argc;
    if (irc_make_chanmode(channel, buffer)) {
	channel->modes = 0;
	buffer[0] = '-';
	irc_mode(oper_numeric(channel), channel, buffer);
	opserv_notice(user, OSMSG_CLEARMODES_DONE, channel->name);
	return 1;
    } else {
	opserv_notice(user, OSMSG_NO_CHANNEL_MODES, channel->name);
	return 1;
    }
}

static OPSERV_FUNC(cmd_deop)
{
    struct userNode *list[256];
    struct modeNode *mn;
    unsigned int count, next;
    const char *numeric;

    OPSERV_MIN_PARMS(2, true);
    numeric = oper_numeric(channel);
    for (count=0, next=1; next<argc; next++) {
	if (!(list[count] = GetUserH(argv[next]))) continue;
        if (IsService(list[count]) && !IsService(user)) continue;
	if (!(mn = GetUserMode(channel, list[count]))) continue;
	if (!(mn->modes & MODE_CHANOP)) continue;
	count++;
	if (count == ArrayLength(list)) {
	    DelChannelOp(count, list, channel, numeric, 1);
	    count = 0;
	}
    }
    if (count) {
	DelChannelOp(count, list, channel, numeric, 1);
    }
    opserv_notice(user, OSMSG_DEOP_DONE);
    return 1;
}

static OPSERV_FUNC(cmd_deopall)
{
    struct modeNode *mn;
    struct userNode *list[256];
    unsigned int count, n;
    const char *numeric;

    OPSERV_NEED_CHANNEL(); (void)argc;
    numeric = oper_numeric(channel);
    for (count=n=0; n<channel->members.used; n++) {
	mn = channel->members.list[n];
	if (IsService(mn->user)
	    || !(mn->modes & MODE_CHANOP)) continue;
	list[count++] = mn->user;
	if (count == ArrayLength(list)) {
	    DelChannelOp(count, list, channel, numeric, 1);
	    count = 0;
	}
    }
    if (count) {
	DelChannelOp(count, list, channel, numeric, 1);
    }
    opserv_notice(user, OSMSG_DEOPALL_DONE, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_rehash)
{
    extern char *services_config;
    (void)channel; (void)argc; (void)argv;

    if(conf_read(services_config))
    {
	opserv_notice(user, OSMSG_REHASH_COMPLETE);
    } else {
	opserv_notice(user, OSMSG_REHASH_FAILED);
    }
    
    return 1;
}

static OPSERV_FUNC(cmd_reopen)
{
    (void)channel; (void)argc; (void)argv;

    reopen_logs();
    opserv_notice(user, OSMSG_REOPEN_COMPLETE);
    
    return 1;
}

static OPSERV_FUNC(cmd_reconnect)
{
    (void)user;(void)channel;(void)argc;(void)argv;

    irc_squit(self, "Reconnecting.");

    return 1;
}

static OPSERV_FUNC(cmd_jupe)
{
    struct server *newsrv;
    unsigned int num;
    char numeric[COMBO_NUMERIC_LEN+1], srvdesc[SERVERDESCRIPTMAX+1];

    OPSERV_MIN_PARMS(4, false); /* name, numeric, desc */
    unsplit_string(argv+3, argc-3);
    num = atoi(argv[2]);
#if USE_UNDERNET
    inttobase64(numeric, num, 2);
    inttobase64(numeric+2, 64*64*64-1, 3);
#else
    if (num < 64) {
        inttobase64(numeric, num, 1);
        inttobase64(numeric+1, 64*64-1, 2);
    } else {
        inttobase64(numeric, num, 2);
        inttobase64(numeric+2, 64*64*64-1, 3);
    }
#endif
    if (GetServer(numeric)) {
        opserv_notice(user, OSMSG_NUMERIC_COLLIDE);
        return 0;
    }
    if (GetServerH(argv[1])) {
        opserv_notice(user, OSMSG_NAME_COLLIDE);
        return 0;
    }
    snprintf(srvdesc, sizeof(srvdesc), "JUPE %s", argv[3]);
    newsrv = AddServer(self, argv[1], 1, now, now, numeric, srvdesc);
    if (!newsrv) {
        opserv_notice(user, OSMSG_SRV_CREATE_FAILED);
        return 0;
    }
    irc_server(newsrv, 0);
    opserv_notice(user, OSMSG_SERVER_JUPED, argv[1]);
    return 1;
}

static OPSERV_FUNC(cmd_unjupe)
{
    struct server *srv;
    OPSERV_MIN_PARMS(2, false);
    srv = GetServerH(argv[1]);
    if (srv) {
        if (strncmp(srv->description, "JUPE", 4)) {
            opserv_notice(user, OSMSG_SERVER_NOT_JUPE);
            return 0;
        } else {
            if (argc > 2) {
                unsplit_string(argv+2, argc-2);
                irc_squit(srv, argv[2]);
            } else {
                irc_squit(srv, "Unjuping server");
            }
            return 1;
        }
    } else {
        opserv_notice(user, MSG_SERVER_UNKNOWN, argv[1]);
        return 0;
    }
}

static OPSERV_FUNC(cmd_jump)
{
    extern struct cManagerNode cManager;
    void uplink_select(char *name);
    int uplink_valid(char *name);

    OPSERV_MIN_PARMS(2, false);
    unsplit_string(argv+1, argc-1);

    if(!strcmp(cManager.uplink->name, argv[1]))
    {
	opserv_notice(user, OSMSG_CURRENT_UPLINK, cManager.uplink->name);
	return 0;
    }

    if(!uplink_valid(argv[1]))
    {
	opserv_notice(user, OSMSG_INVALID_UPLINK, argv[1]);
	return 0;
    }

    uplink_select(argv[1]);
    irc_squit(self, "Reconnecting.");

    return 1;
}

static OPSERV_FUNC(cmd_die)
{
    char *reason;

    OPSERV_MIN_PARMS(2, false);
    unsplit_string(argv+1, argc-1);

    reason = alloca(strlen(argv[1]) + strlen(user->nick) + 20);
    sprintf(reason, "Disconnected by %s [%s]", user->nick, argv[1]);

    irc_squit(self, reason);

    quit_services = 1;
    return 1;
}

static OPSERV_FUNC(cmd_restart)
{
    extern int services_argc;
    extern char **services_argv;
    char **restart_argv, *reason;

    OPSERV_MIN_PARMS(2, false);
    unsplit_string(argv+1, argc-1);

    reason = alloca(strlen(argv[1]) + strlen(user->nick) + 17);
    sprintf(reason, "Restarted by %s [%s]", user->nick, argv[1]);

    irc_squit(self, reason);

    /* Append a NULL to the end of argv[]. */
    restart_argv = (char **)alloca((services_argc + 1) * sizeof(char *));
    memcpy(restart_argv, services_argv, services_argc * sizeof(char *));
    restart_argv[services_argc] = NULL;

    call_exit_funcs();

    /* Don't blink. */
    execv(services_argv[0], restart_argv);

    /* If we're still here, that means something went wrong. Reconnect. */
    return 1;
}

static void
opserv_block(struct userNode *target, char *src_handle, char *reason)
{
    int bsize;
    char *mask;
    bsize = 3+strlen(target->hostname);
    mask = alloca(bsize);
    snprintf(mask, bsize, "*@%s", target->hostname);
    if (!reason) {
        bsize = strlen(OSMSG_GLINE_REQUESTED)+strlen(src_handle)+1;
        reason = alloca(bsize);
        snprintf(reason, bsize, OSMSG_GLINE_REQUESTED, src_handle);
    }
    gline_add(src_handle, mask, 3600, reason, 1);
}

static OPSERV_FUNC(cmd_block)
{
    struct userNode *target;

    OPSERV_MIN_PARMS(2, false);
    target = GetUserH(argv[1]);
    if (!target) {
	opserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	return 0;
    }
    if (IsService(target)) {
	opserv_notice(user, MSG_SERVICE_IMMUNE, target->nick);
	return 0;
    }
    if (argc > 2) {
        unsplit_string(argv+2, argc-2);
        opserv_block(target, user->handle_info->handle, argv[2]);
    } else {
        opserv_block(target, user->handle_info->handle, NULL);
    }
    opserv_notice(user, OSMSG_GLINE_ISSUED, target->hostname);
    return 1;
}

static OPSERV_FUNC(cmd_gline)
{
    unsigned long duration;
    OPSERV_MIN_PARMS(4, false);
    unsplit_string(argv+3, argc-3);

    if (!is_gline(argv[1])) {
	opserv_notice(user, MSG_INVALID_GLINE, argv[1]);
	return 0;
    }
    duration = ParseInterval(argv[2]);
    if (!duration) {
        opserv_notice(user, MSG_INVALID_DURATION, argv[2]);
        return 0;
    }
    gline_remove(argv[1], 0);
    gline_add(user->handle_info->handle, argv[1], duration, argv[3], 1);
    opserv_notice(user, OSMSG_GLINE_ISSUED, argv[1]);
    return 1;
}

static OPSERV_FUNC(cmd_ungline)
{
    OPSERV_MIN_PARMS(2, false);
    gline_remove(argv[1], 1);
    opserv_notice(user, OSMSG_GLINE_REMOVED, argv[1]);
    return 1;
}

static OPSERV_FUNC(cmd_refreshg)
{
    (void)channel; (void)argc; (void)argv;
    gline_refresh_all();
    opserv_notice(user, OSMSG_GLINES_REFRESHED);
    return 1;
}

static void
opserv_ison(struct userNode *tell, struct userNode *target, const char *prefix)
{
    struct modeNode *mn;
    unsigned int count, here_len, n;
    char buff[460];

    count = snprintf(buff, sizeof(buff), prefix, target->nick);
    for (n=0; n<target->channels.used; n++) {
	mn = target->channels.list[n];
	here_len = strlen(mn->channel->name);
	if ((count + here_len + 4) > sizeof(buff)) {
	    buff[count] = 0;
	    opserv_notice(tell, "%s", buff);
	    count = 0;
	}
	buff[count++] = ' ';
	if (mn->modes & MODE_CHANOP) buff[count++] = '@';
	if (mn->modes & MODE_VOICE) buff[count++] = '+';
	memcpy(buff+count, mn->channel->name, here_len);
	count += here_len;
    }
    if (count) {
	buff[count] = 0;
	opserv_notice(tell, "%s", buff);
    }
}

static OPSERV_FUNC(cmd_inviteme)
{
    struct userNode *target;
    (void)channel;
    if (argc < 2) {
	target = user;
    } else {
	target = GetUserH(argv[1]);
	if (!target) {
	    opserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	    return 0;
	}
    }
    if (opserv_conf.debug_channel == NULL) {
	opserv_notice(user, OSMSG_NO_DEBUG_CHANNEL);
	return 0;
    }
    irc_invite(opserv->numeric, target, opserv_conf.debug_channel);
    if (target != user) {
	opserv_notice(user, OSMSG_INVITE_DONE, target->nick, opserv_conf.debug_channel->name);
    }
    return 1;
}

static OPSERV_FUNC(cmd_join)
{
    (void)argc;
    if (!channel) {
	if (!argv[-1]) {
	    opserv_notice(user, MSG_MISSING_PARAMS, argv[0]);
	    OPSERV_SYNTAX();
	    return 0;
	}
	channel = AddChannel(argv[-1], now, "", "", 0, opserv->numeric, "");
	AddChannelOp(1, &opserv, channel, opserv->numeric, 0);
    } else {
	AddChannelUser(opserv, channel, 0);
	AddChannelOp(1, &opserv, channel, self->numeric, 1);
    }
    irc_fetchtopic(opserv->numeric, channel->name);
    opserv_notice(user, OSMSG_JOIN_DONE, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_ison)
{
    struct userNode *target;

    OPSERV_MIN_PARMS(2, false);
    target = GetUserH(argv[1]);
    if (!target) {
	opserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	return 0;
    }
    opserv_ison(user, target, OSMSG_ISON_PREFIX);
    return 1;
}

static OPSERV_FUNC(cmd_kick)
{
    struct userNode *target;
    char *reason;

    OPSERV_MIN_PARMS(2, true);
    if (argc < 3) {
	reason = alloca(strlen(OSMSG_KICK_REQUESTED)+strlen(user->nick)+1);
	sprintf(reason, OSMSG_KICK_REQUESTED, user->nick);
    } else {
	unsplit_string(argv+2, argc-2);
	reason = argv[2];
    }
    target = GetUserH(argv[1]);
    if (!target) {
	opserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	return 0;
    }
    if (!GetUserMode(channel, target)) {
	opserv_notice(user, OSMSG_NOT_ON_CHANNEL, target->nick, channel->name);
	return 0;
    }
    KickChannelUser(target, channel, opserv, reason);
    return 1;
}

static OPSERV_FUNC(cmd_kickall)
{
    unsigned int limit;
    unsigned int n, inchan;
    struct modeNode *mn;
    char *reason;

    OPSERV_NEED_CHANNEL();
    /* ircu doesn't let servers KICK users, so if OpServ's not in the
     * channel, we have to join it in temporarily. */
    if (!(inchan = GetUserMode(channel, opserv) ? 1 : 0)) {
        AddChannelUser(opserv, channel, 0);
	AddChannelOp(1, &opserv, channel, self->numeric, 1);
    }
    if (argc < 2) {
	reason = alloca(strlen(OSMSG_KICK_REQUESTED)+strlen(user->nick)+1);
	sprintf(reason, OSMSG_KICK_REQUESTED, user->nick);
    } else {
	unsplit_string(argv+1, argc-1);
	reason = argv[1];
    }
    limit = user->handle_info->opserv_level;
    for (n=channel->members.used; n>0;) {
	mn = channel->members.list[--n];
	if (IsService(mn->user)
	    || (mn->user->handle_info
		&& (mn->user->handle_info->opserv_level >= limit))) {
	    continue;
	}
	KickChannelUser(mn->user, channel, opserv, reason);
    }
    if (!inchan) {
        DelChannelUser(opserv, channel, reason, 0);
    }
    opserv_notice(user, OSMSG_KICKALL_DONE, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_kickban)
{
    struct userNode *target;
    char *reason;
    char *mask;

    OPSERV_MIN_PARMS(2, true);
    if (argc == 2) {
	reason = alloca(strlen(OSMSG_KICK_REQUESTED)+strlen(user->nick)+1);
	sprintf(reason, OSMSG_KICK_REQUESTED, user->nick);
    } else {
	unsplit_string(argv+2, argc-2);
	reason = argv[2];
    }
    target = GetUserH(argv[1]);
    if (!target) {
	opserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	return 0;
    }
    if (!GetUserMode(channel, target)) {
	opserv_notice(user, OSMSG_NOT_ON_CHANNEL, target->nick, channel->name);
	return 0;
    }
    mask = generate_hostmask(target, 0);
    AddChannelBan(1, &mask, channel, oper_numeric(channel), time(0), 1);
    free(mask);
    KickChannelUser(target, channel, opserv, reason);
    return 1;
}

static OPSERV_FUNC(cmd_kickbanall)
{
    unsigned int limit, m, count;
    int n;
    struct modeNode *mn;
    struct banNode *bn;
    char **list, *reason;

    OPSERV_NEED_CHANNEL(); (void)argc;
    if (argc < 2) {
	reason = alloca(strlen(OSMSG_KICK_REQUESTED)+strlen(user->nick)+1);
	sprintf(reason, OSMSG_KICK_REQUESTED, user->nick);
    } else {
	unsplit_string(argv+1, argc-1);
	reason = argv[1];
    }
    list = alloca(channel->members.used*sizeof(list[0]));
    limit = user->handle_info->opserv_level;
    /* ban them first */
    for (count=n=0; n<(signed)channel->members.used; n++) {
	mn = channel->members.list[n];
	if (IsService(mn->user)
	    || (mn->user->handle_info
		&& (mn->user->handle_info->opserv_level >= limit))) {
	    continue;
	}
	/* skip this user if they match a pre-existing ban */
	for (m=0; m<channel->banlist.used; m++) {
	    bn = channel->banlist.list[m];
	    if (user_matches_glob(mn->user, bn->ban, 1)) break;
	}
	if (m<channel->banlist.used) continue;
	/* see if they match a ban we've just constructed */
	for (m=0; m<count; m++) {
	    if (user_matches_glob(mn->user, list[m], 1)) break;
	}
	if (m<count) continue;
	list[count++] = generate_hostmask(mn->user, 0);
    }
    if (count) {
	AddChannelBan(count, list, channel, oper_numeric(channel), time(0), 1);
    }
    /* now kick them */
    for (n=channel->members.used-1; n>=0; n--) {
	mn = channel->members.list[n];
	if (IsService(mn->user)
	    || (mn->user->handle_info
		&& (mn->user->handle_info->opserv_level >= limit))) {
	    continue;
	}
	KickChannelUser(mn->user, channel, opserv, reason);
    }
    opserv_notice(user, OSMSG_KICKALL_DONE, channel->name);
    return 1;    
}

static OPSERV_FUNC(cmd_part)
{
    OPSERV_NEED_CHANNEL(); (void)argc;
    DelChannelUser(opserv, channel, "Leaving.", 0);
    opserv_notice(user, OSMSG_LEAVE_DONE, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_mode)
{
    const char *numeric;

    OPSERV_MIN_PARMS(2, true);
    unsplit_string(argv+1, argc-1);
    numeric = oper_numeric(channel);
    if (channel->modes & MODE_KEY) {
	char *mc = alloca(strlen(channel->key)+4);
	sprintf(mc, "-k %s", channel->key);
	irc_mode(numeric, channel, mc);
    }
    mod_chanmode(channel, argv[1], NULL, 0);
    irc_mode(numeric, channel, argv[1]);
    opserv_notice(user, OSMSG_MODE_SET, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_nickban)
{
    struct userNode *target;
    char *usermask;
    OPSERV_MIN_PARMS(2, true);
    target = GetUserH(argv[1]);
    if (!target) {
	opserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	return 0;
    }
    usermask = generate_hostmask(target, 0);
    AddChannelBan(1, &usermask, channel, oper_numeric(channel), now, 1);
    free(usermask);
    opserv_notice(user, OSMSG_ADDED_BAN, usermask, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_op)
{
    struct userNode *list[256];
    struct modeNode *mn;
    const char *numeric;
    unsigned int count, next;

    OPSERV_MIN_PARMS(2, true);
    numeric = oper_numeric(channel);
    for (count=0, next=1; next<argc; next++) {
	if (!(list[count] = GetUserH(argv[next]))) continue;
	if (!(mn = GetUserMode(channel, list[count]))) continue;
	if (mn->modes & MODE_CHANOP) continue;
	count++;
	if (count == ArrayLength(list)) {
	    AddChannelOp(count, list, channel, numeric, 1);
	    count = 0;
	}
    }
    if (count) {
	AddChannelOp(count, list, channel, numeric, 1);
    }
    opserv_notice(user, OSMSG_OP_DONE, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_opall)
{
    struct userNode *users[256];
    struct modeNode *mn;
    unsigned int count, n;
    const char *numeric;

    (void)argv;
    OPSERV_NEED_CHANNEL(); (void)argc;
    numeric = oper_numeric(channel);
    for (count=n=0; n<channel->members.used; n++) {
	mn = channel->members.list[n];
	if (mn->modes & MODE_CHANOP) continue;
	users[count++] = mn->user;
	if (count == ArrayLength(users)) {
	    AddChannelOp(count, users, channel, numeric, 1);
	    count = 0;
	}
    }
    if (count) {
	AddChannelOp(count, users, channel, numeric, 1);
    }
    opserv_notice(user, OSMSG_OP_DONE, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_whois)
{
    struct userNode *target;
    unsigned long ip;
    char buffer[128];
    int bpos, herelen;

    OPSERV_MIN_PARMS(2, false);
    target = GetUserH(argv[1]);
    if (!target) {
	opserv_notice(user, MSG_NICK_UNKNOWN, argv[1]);
	return 0;
    }
    ip = target->ip;
    opserv_notice(user, OSMSG_WHOIS_NICK, target->nick);
    opserv_notice(user, OSMSG_WHOIS_HOST, target->username, target->hostname);
    opserv_notice(user, OSMSG_WHOIS_IP, (ip >> 24) & 255, (ip >> 16) & 255, (ip >> 8) & 255, ip & 255);
    if (target->modes) {
	bpos = 0;
#define buffer_cat(str) (herelen = strlen(str), memcpy(buffer+bpos, str, herelen), bpos += herelen)
	buffer_cat("Modes   : +");
	if (IsInvisible(target)) buffer_cat("i");
	if (IsWallOp(target)) buffer_cat("w");
	if (IsOper(target)) buffer_cat("o");
	if (IsGlobal(target)) buffer_cat("g");
	if (IsServNotice(target)) buffer_cat("s");
	if (IsHelperIrcu(target)) buffer_cat("h");
	if (IsService(target)) buffer_cat("k");
	if (IsDeaf(target)) buffer_cat("d");
        if (IsGagged(target)) buffer_cat("; gagged");
	buffer[bpos] = 0;
	if (buffer[11]) opserv_notice(user, buffer);
    }
    opserv_notice(user, OSMSG_WHOIS_INFO, target->info);
    opserv_notice(user, OSMSG_WHOIS_NUMERIC, target->numeric);
    opserv_notice(user, OSMSG_WHOIS_SERVER, target->uplink->name);
    if (target->channels.used <= MAX_CHANNELS_WHOIS) {
	opserv_ison(user, target, OSMSG_WHOIS_CHANNELS);
    } else {
	opserv_notice(user, OSMSG_WHOIS_HIDECHANS);
    }
    return 1;
}

static OPSERV_FUNC(cmd_unban)
{
    OPSERV_MIN_PARMS(2, true);
    DelChannelBan(argc-1, (char**)argv+1, channel, oper_numeric(channel), 1);
    opserv_notice(user, OSMSG_UNBAN_DONE, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_moderate)
{
    struct userNode *users[256];
    unsigned int n, count;
    struct modeNode *mn;
    const char *numeric;
    (void)argv;
    OPSERV_NEED_CHANNEL(); (void)argc;
    if (channel->modes & MODE_MODERATED) {
	opserv_notice(user, OSMSG_ALREADY_MODERATED, channel->name);
	return 0;
    }
    numeric = oper_numeric(channel);
    for (count=n=0; n<channel->members.used; n++) {
	mn = channel->members.list[n];
	if (mn->modes & MODE_CHANOP) continue;
	users[count++] = mn->user;
	if (count == ArrayLength(users)) {
	    AddChannelVoice(count, users, channel, numeric, 1);
	    count = 0;
	}
    }
    if (count) {
	AddChannelVoice(count, users, channel, numeric, 1);
    }
    irc_mode(numeric, channel, "+m");
    channel->modes |= MODE_MODERATED;
    opserv_notice(user, OSMSG_CHANNEL_MODERATED, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_unmoderate)
{
    (void)argv;
    OPSERV_NEED_CHANNEL(); (void)argc;
    if (!(channel->modes & MODE_MODERATED)) {
	opserv_notice(user, OSMSG_ALREADY_UNMODERATED, channel->name);
	return 0;
    }
    irc_mode(oper_numeric(channel), channel, "-m");
    channel->modes &= ~MODE_MODERATED;
    opserv_notice(user, OSMSG_CHANNEL_UNMODERATED, channel->name);
    return 1;
}

static OPSERV_FUNC(cmd_chlevel)
{
    struct opserv_cmd *cmd;
    int level;

    OPSERV_MIN_PARMS(3, false);

    unsplit_string(argv+1, argc-2);
    cmd = dict_find(opserv_func_dict, argv[1], NULL);
    if (!cmd) {
	opserv_notice(user, MSG_COMMAND_UNKNOWN, argv[1]);
	return 0;
    }
    level = strtoul(argv[argc-1], NULL, 0);
    cmd->min_level = level;
    opserv_notice(user, OSMSG_COMMAND_ACCESS_IS, argv[1], level);
    return 1;
}

static OPSERV_FUNC(cmd_cmdlevel)
{
    struct opserv_cmd *cmd;

    OPSERV_MIN_PARMS(2, false);

    unsplit_string(argv+1, argc-1);
    cmd = dict_find(opserv_func_dict, argv[1], NULL);
    if (cmd) {
        opserv_notice(user, OSMSG_COMMAND_ACCESS_IS, argv[1], cmd->min_level);
        return 1;
    } else {
	opserv_notice(user, MSG_COMMAND_UNKNOWN, argv[1]);
	return 0;
    }
}

static OPSERV_FUNC(cmd_enablecom)
{
    struct opserv_cmd *cmd;

    OPSERV_MIN_PARMS(2, false);
    cmd = dict_find(opserv_func_dict, argv[1], NULL);
    if (!cmd) {
	opserv_notice(user, MSG_COMMAND_UNKNOWN, argv[1]);
	return 0;
    }
    cmd->disabled = 0;
    opserv_notice(user, OSMSG_COMMAND_ENABLED, argv[1]);
    return 1;
}

static OPSERV_FUNC(cmd_disablecom)
{
    struct opserv_cmd *cmd;

    OPSERV_MIN_PARMS(2, false);
    cmd = dict_find(opserv_func_dict, argv[1], NULL);
    if (!cmd) {
	opserv_notice(user, MSG_COMMAND_UNKNOWN, argv[1]);
	return 0;
    } else if (cmd->func == cmd_enablecom) {
	/* Stupid as it may be, it would really suck if someone did this. */
	opserv_notice(user, OSMSG_DUMBASS, argv[1]);
	return 0;
    }
    cmd->disabled = 1;
    opserv_notice(user, MSG_COMMAND_DISABLED, argv[1]);
    return 1;
}

static OPSERV_FUNC(opserv_missing_support)
{
    OPSERV_MIN_PARMS(1, false);
    opserv_notice(user, OSMSG_MISSING_SUPPORT);
    return 1;
}

static int
append_if_disabled(const char *key, void *data, void *extra)
{
    struct opserv_cmd *cmd = data;
    if (cmd->disabled) string_list_append(extra, strdup(key));
    return 0;
}

static OPSERV_FUNC(cmd_showdiscom)
{
    char buffer[400];
    unsigned int i;
    int start, here_len;
    struct string_list *strlist;

    (void)argc; (void)argv; (void)channel;
    strlist = alloc_string_list(5);
    dict_foreach(opserv_func_dict, append_if_disabled, strlist);
    start = sizeof(buffer)-1;
    buffer[start] = 0;
    for (i=0; i<strlist->used; i++) {
	here_len = strlen(strlist->list[i]);
	if (start-here_len-1 < 0) {
	    opserv_notice(user, OSMSG_DISABLED_LIST, buffer+start);
	    start = sizeof(buffer)-1;
	}
	start -= here_len;
	memcpy(buffer+start, strlist->list[i], here_len); /* omit terminator */
	buffer[--start] = ' ';
    }
    if (start != sizeof(buffer)-1) {
	opserv_notice(user, OSMSG_DISABLED_LIST, buffer+start);
    }
    free_string_list(strlist);
    return 1;
}

static void
cmd_stats_bad(struct userNode *user)
{
    char buffer[400];
    unsigned int i;
    int start, here_len;
    start = sizeof(buffer)-1;
    buffer[start] = 0;
    for (i=0; i<opserv_bad_words->used; i++) {
	here_len = strlen(opserv_bad_words->list[i]);
	if (start-here_len-1 < 0) {
	    opserv_notice(user, OSMSG_BADWORD_LIST, buffer+start);
	    start = sizeof(buffer)-1;
	}
	start -= here_len;
	memcpy(buffer+start, opserv_bad_words->list[i], here_len); /* omit terminator */
	buffer[--start] = ' ';
    }
    if (start != sizeof(buffer)-1) {
	opserv_notice(user, OSMSG_BADWORD_LIST, buffer+start);
    }
}

static void
cmd_stats_glines(struct userNode *user)
{
    opserv_notice(user, OSMSG_GLINE_COUNT, gline_count());
}

static void
trace_links(struct userNode *user, struct server *server, int allow_self)
{
    unsigned int n;

    if(!allow_self && server == self) {
	return;
    }

    for (n=0; n < ArrayLength(servers); n++) {
	if (!servers[n] || (servers[n]->uplink != server)) continue;
	opserv_notice(user, OSMSG_LINK_INFO, servers[n]->name, server->name);

	trace_links(user, servers[n], 0);
    }
}

static void
cmd_stats_links(struct userNode *user)
{
    opserv_notice(user, OSMSG_BEGIN_LINKS);
    trace_links(user, self, 1);
    opserv_notice(user, OSMSG_END_LINKS);
}

static void
cmd_stats_max(struct userNode *user)
{
    opserv_notice(user, OSMSG_MAX_CLIENTS, max_clients, asctime(localtime(&max_clients_time)));
}

static void
cmd_stats_network(struct userNode *user)
{
    unsigned int n, tot_clients;

    tot_clients = dict_size(clients);
    opserv_notice(user, OSMSG_NETWORK_INFO, tot_clients, invis_clients, curr_opers.used);
    for (n=0; n<ArrayLength(servers); n++) {
	if (!servers[n]) continue;
	opserv_notice(user, OSMSG_NETWORK_SERVER,
		      servers[n]->name,
		      servers[n]->clients,
		      ((double)servers[n]->clients/tot_clients)*100);
    }
}

static void
cmd_stats_network_2(struct userNode *user)
{
    unsigned int n;
    char uptime_str[INTERVALLEN];

    for (n=0; n<ArrayLength(servers); n++) {
	if (!servers[n]) continue;
        intervalString(uptime_str, now - servers[n]->link);
	opserv_notice(user, OSMSG_NETWORK_SERVER_2,
		      servers[n]->name,
		      servers[n]->numeric,
		      base64toint(servers[n]->numeric, strlen(servers[n]->numeric)),
                      uptime_str);
    }
}


static void
cmd_stats_opers(struct userNode *user)
{
    unsigned int n;
    opserv_notice(user, OSMSG_STATS_OPER_LIST);
    for (n=0; n<curr_opers.used; n++) {
	struct userNode *oper = curr_opers.list[n];
	opserv_notice(user, IDENT_FORMAT, IDENT_DATA(oper));
    }
}

static int
reserved_helper(const char *nick, void *data, void *extra)
{
    struct userNode *user = extra;
    (void)data;
    opserv_notice(user, OSMSG_RESERVED_NICK_ITEM, nick);
    return 0;
}

static void
cmd_stats_reserved(struct userNode *user)
{
    opserv_notice(user, OSMSG_RESERVED_LIST);
    dict_foreach(opserv_reserved_nick_dict, reserved_helper, user);
}

static void
cmd_stats_secret(struct userNode *user)
{
    unsigned int n;
    opserv_notice(user, OSMSG_SECRET_LIST);
    for (n=0; n<opserv_secret_words->used; n++) {
	opserv_notice(user, OSMSG_WORD_IS_SECRET, opserv_secret_words->list[n]);
    }
}

static int
cmd_stats_trusted_helper(const char *key, void *data, void *extra)
{
    struct trusted_host *th;
    char length[INTERVALLEN];

    th = data;
    if (th->expires) intervalString(length, th->expires - now);
    opserv_notice(extra, OSMSG_HOST_IS_TRUSTED, key, th->limit, (th->expires ? length : "never"), th->reason);
    return 0;
}

static void
cmd_stats_trusted(struct userNode *user)
{
    opserv_notice(user, OSMSG_TRUSTED_LIST);
    dict_foreach(opserv_trusted_hosts, cmd_stats_trusted_helper, user);
}

static void
cmd_stats_uptime(struct userNode *user)
{
    char uptime[INTERVALLEN];
    struct tms buf;
    extern time_t boot_time;
    extern int lines_processed;

    intervalString(uptime, now-boot_time);
    times(&buf);
    opserv_notice(user, OSMSG_UPTIME_STATS,
		  uptime, lines_processed,
		  buf.tms_utime/(double)CLOCKS_PER_SEC,
		  buf.tms_stime/(double)CLOCKS_PER_SEC);
    if (last_db_write) {
        intervalString(uptime, now - last_db_write);
        opserv_notice(user, OSMSG_LAST_WRITE, uptime);
    } else {
        opserv_notice(user, OSMSG_NO_LAST_WRITE);
    }
}

static void
cmd_stats_proxycheck(struct userNode *user)
{
    struct sockcheck_stats stats;
    sockcheck_get_stats(&stats);
    opserv_notice(user, OSMSG_SOCKCHECK_STATS, stats.checked_ip_count,
		  stats.failed_ip_count, stats.checking_ip_count,
		  stats.checking_ip_max, stats.pending_ip_count,
		  stats.cached_ip_count, stats.test_count);
}

static int
cmd_stats_alerts_helper(const char *key, void *data, void *extra)
{
    struct opserv_user_alert *alert = data;
    const char *reaction;
    switch (alert->reaction) {
    case REACT_NOTICE: reaction = "notice"; break;
    case REACT_GLINE: reaction = "gline"; break;
    default: reaction = "<unknown>"; break;
    }
    opserv_notice(extra, OSMSG_ALERT_IS, key, alert->owner, reaction, alert->text_discrim);
    return 0;
}

static void
cmd_stats_alerts(struct userNode *user)
{
    opserv_notice(user, OSMSG_ALERTS_LIST);
    dict_foreach(opserv_user_alerts, cmd_stats_alerts_helper, user);
}

static void
cmd_stats_gags(struct userNode *user)
{
    if (!gagList) {
	opserv_notice(user, OSMSG_NO_GAGS);
    } else {
        struct gag_entry *gag;
        opserv_notice(user, OSMSG_GAG_HEADER);
        for (gag = gagList; gag; gag = gag->next) {
            char expstr[INTERVALLEN];
            if (gag->expires) intervalString(expstr, gag->expires-now);
            opserv_notice(user, OSMSG_GAG_DATA, gag->mask, gag->owner, (gag->expires?expstr:"Never"), gag->reason);
        }
    }
}

static void
cmd_stats_timeq(struct userNode *user)
{
    opserv_notice(user, OSMSG_TIMEQ_INFO, timeq_size(), timeq_next()-now);
}

static OPSERV_FUNC(cmd_stats)
{
    OPSERV_MIN_PARMS(2, false);
    if (!strcasecmp(argv[1], "bad")) cmd_stats_bad(user);
    else if (!strcasecmp(argv[1], "glines")) cmd_stats_glines(user);
    else if (!strcasecmp(argv[1], "links")) cmd_stats_links(user);
    else if (!strcasecmp(argv[1], "max")) cmd_stats_max(user);
    else if (!strcasecmp(argv[1], "network")) cmd_stats_network(user);
    else if (!strcasecmp(argv[1], "network2")) cmd_stats_network_2(user);
    else if (!strcasecmp(argv[1], "opers")) cmd_stats_opers(user);
    else if (!strcasecmp(argv[1], "reserved")) cmd_stats_reserved(user);
    else if (!strcasecmp(argv[1], "secret")) cmd_stats_secret(user);
    else if (!strcasecmp(argv[1], "trusted")) cmd_stats_trusted(user);
    else if (!strcasecmp(argv[1], "uptime")) cmd_stats_uptime(user);
    else if (!strcasecmp(argv[1], "alerts")) cmd_stats_alerts(user);
    else if (!strcasecmp(argv[1], "gags")) cmd_stats_gags(user);
    else if (!strcasecmp(argv[1], "timeq")) cmd_stats_timeq(user);
    else if (!strcasecmp(argv[1], "proxycheck")) {
        if (SOCKCHECK_USABLE) {
            cmd_stats_proxycheck(user);
        } else {
            opserv_missing_support(user, channel, argc, argv);
        }
    } else {
	opserv_notice(user, OSMSG_UNKNOWN_STATS, argv[1]);
	OPSERV_SYNTAX();
	return 0;
    }
    return 1;
}

static OPSERV_FUNC(cmd_hostscan)
{
    unsigned int n;
    unsigned long addr;

    OPSERV_MIN_PARMS(2, false);
    for (n=1; n<argc; n++) {
	struct userNode *un = GetUserH(argv[n]);
	char *scanhost;

	if (!un) {
	    scanhost = argv[n];
	} else {
	    scanhost = un->hostname;
	}

 	if (getipbyname(scanhost, &addr)) {
	    sockcheck_queue_address(addr);
	    opserv_notice(user, OSMSG_ADDRESS_QUEUED, scanhost);
	} else {
	    opserv_notice(user, OSMSG_NO_SUCH_ADDRESS, scanhost);
	}
    }
    return 1;
}

static OPSERV_FUNC(cmd_clearhost)
{
    unsigned int n, i;

    OPSERV_MIN_PARMS(2, false);

    for (n=1; n<argc; n++) {
        struct userNode *un = GetUserH(argv[n]);
	char *scanhost;

        if (!un) {
	    scanhost = argv[n];
        } else {
	    scanhost = un->hostname;
        }

	if ((i = sockcheck_uncache_host(scanhost)) && i != (unsigned)-1) {
	    opserv_notice(user, OSMSG_HOST_CLEARED, scanhost);
	} else if (i == (unsigned)-1) {
	    opserv_notice(user, OSMSG_HOST_BUSY, scanhost);
	} else {
	    opserv_notice(user, OSMSG_HOST_NOTCLEARED, scanhost);
	}
    }
    return 1;
}

static OPSERV_FUNC(cmd_dump)
{
    char *linedup;
    OPSERV_MIN_PARMS(2, false);

    unsplit_string(argv+1, argc-1);
    /* assume it's only valid IRC if we can parse it */
    linedup = alloca(strlen(argv[1])+1);
    strcpy(linedup, argv[1]);
    if (parse_line(linedup, 1)) {
	irc_raw(argv[1]);
	opserv_notice(user, OSMSG_LINE_DUMPED);
    } else {
	opserv_notice(user, OSMSG_RAW_PARSE_ERROR);
    }
    return 1;
}

static OPSERV_FUNC(cmd_raw)
{
    char *linedup;
    OPSERV_MIN_PARMS(2, false);

    unsplit_string(argv+1, argc-1);
    /* send it anyways, and parse it as well 
       if something is wrong, hopefully it will
       core services before it cores a hub      */
    linedup = alloca(strlen(argv[1])+1);
    strcpy(linedup, argv[1]);
    parse_line(linedup, 1);
    irc_raw(argv[1]);
    opserv_notice(user, OSMSG_LINE_DUMPED);
    return 1;
}

struct userNode *
opserv_add_reserve(struct userNode *user, const char *nick, const char *username, const char *host, const char *desc)
{
    struct userNode *resv = GetUserH(nick);
    if (resv) {
	if (IsService(resv)) {
	    opserv_notice(user, MSG_SERVICE_IMMUNE, resv->nick);
	    return NULL;
	}
	if (resv->handle_info
	    && resv->handle_info->opserv_level > user->handle_info->opserv_level) {
	    opserv_notice(user, OSMSG_LEVEL_TOO_LOW);
	    return NULL;
	}
	DelUser(resv, opserv, true, OSMSG_NICK_RESERVED);
    }
    resv = AddClone(nick, username, host, desc);
    dict_insert(opserv_reserved_nick_dict, resv->nick, resv);
    return resv;
}

static OPSERV_FUNC(cmd_collide)
{
    struct userNode *resv;

    OPSERV_MIN_PARMS(5, false);

    unsplit_string(argv+4, argc-4);
    resv = opserv_add_reserve(user, argv[1], argv[2], argv[3], argv[4]);
    if (resv) {
	opserv_notice(user, OSMSG_COLLIDED_NICK, resv->nick);
	return 1;
    } else {
	return 0;
    }
}

static OPSERV_FUNC(cmd_reserve)
{
    struct userNode *resv;

    OPSERV_MIN_PARMS(5, false);

    unsplit_string(argv+4, argc-4);
    resv = opserv_add_reserve(user, argv[1], argv[2], argv[3], argv[4]);
    if (resv) {
	resv->modes |= FLAGS_PERSISTENT;
	opserv_notice(user, OSMSG_RESERVED_NICK, resv->nick);
	return 1;
    } else {
	return 0;
    }
}

static int
free_reserve(char *nick)
{
    struct userNode *resv;
    unsigned int rlen;
    char *reason;

    resv = dict_find(opserv_reserved_nick_dict, nick, NULL);
    if(!resv) return 0;

    rlen = strlen(resv->nick)+strlen(OSMSG_PART_REASON);
    reason = alloca(rlen);
    snprintf(reason, rlen, OSMSG_PART_REASON, resv->nick);
    DelUser(resv, opserv, true, reason);

    dict_remove(opserv_reserved_nick_dict, nick);
    return 1;
}

static OPSERV_FUNC(cmd_unreserve)
{
    OPSERV_MIN_PARMS(2, false);

    if (free_reserve(argv[1])) {
	opserv_notice(user, OSMSG_NICK_UNRESERVED, argv[1]);
    } else {
	opserv_notice(user, OSMSG_NOT_RESERVED, argv[1]);
    }
    return 1;
}

static OPSERV_FUNC(cmd_defproxy)
{
    const char *reason;

    OPSERV_MIN_PARMS(2, false);

    unsplit_string(argv+1, argc-1);
    if ((reason = sockcheck_add_test(argv[1]))) {
	opserv_notice(user, OSMSG_DEFPROXY_FAILED, reason);
	return 0;
    }
    opserv_notice(user, OSMSG_DEFPROXY_SUCCESS);
    return 1;
}

static void
opserv_part_channel(void *data)
{
    struct chanNode *channel = data;
    DelChannelUser(opserv, channel, "Leaving.", 0);
}

static int
opserv_trusted_host_check(struct userNode *user)
{
    struct opserv_hostinfo *ohi;
    if (!(ohi = dict_find(opserv_hostinfo_dict, user->hostname, NULL))) {
	ohi = malloc(sizeof(*ohi));
	ohi->client_count = 0;
	dict_insert(opserv_hostinfo_dict, strdup(user->hostname), ohi);
    }
    ohi->client_count++;
    if (!opserv_conf.untrusted_max) return 0;
    if (ohi->client_count == opserv_conf.untrusted_max) {
        opserv_notice(user, OSMSG_CLONE_WARNING);
    } else if (ohi->client_count > opserv_conf.untrusted_max) {
	struct trusted_host *th = dict_find(opserv_trusted_hosts, user->hostname, NULL);
	if (!th || (ohi->client_count > th->limit)) {
	    char *target = alloca(3+strlen(user->hostname));
	    sprintf(target, "*@%s", user->hostname);
	    gline_add(opserv->nick, target, 3600, OSMSG_TOO_MANY_CLONES, 1);
	}
    }
    return 0;
}

static void
opserv_user_cleanup(struct userNode *user)
{
    struct opserv_hostinfo *ohi;

    /* Any data that might be associated with users within
       external structures should be cleaned up here. */
    if (user->uplink == self)
    {
	/* Try to remove it from the reserved nick dict without calling
	   free_reserve, because that would call DelUser().. */
	dict_remove(opserv_reserved_nick_dict, user->nick);
	return;
    }
    if ((ohi = dict_find(opserv_hostinfo_dict, user->hostname, NULL))) {
	ohi->client_count--;
	if (ohi->client_count == 0) dict_remove(opserv_hostinfo_dict, user->hostname);
    }
}

static int
opserv_proxy_check(struct userNode *user)
{
    if (user->ip) {
        sockcheck_queue_address(htonl(user->ip));
    }
    return 0;
}

static int
opserv_nick_flood_check(struct userNode *user)
{
    (void)user;
    if (!policer_conforms(opserv_conf.new_user_policer, now, 1.0)) {
	if (!new_user_flood) {
	    new_user_flood = 1;
	    opserv_alert(OSMSG_NEW_USER_FLOOD);
	}
    } else {
	new_user_flood = 0;
    }
    return 0;
}

static int
opserv_bad_channel(const char *name)
{
    unsigned int n;

    for (n=0; n<opserv_bad_words->used; n++) {
	if (strcasestr(name, opserv_bad_words->list[n])) {
	    return 1;
	}
    }

    return 0;
}

static void
opserv_shutdown_channel(struct chanNode *channel)
{
    unsigned int nn;

    AddChannelUser(opserv, channel, 0);
    AddChannelOp(1, &opserv, channel, self->numeric, 1);
    channel->modes |= MODE_SECRET | MODE_INVITEONLY;
    irc_mode(self->numeric, channel, "+sib *!*@*");
    timeq_add(now + opserv_conf.purge_lock_delay, opserv_part_channel, channel);

    for (nn=channel->members.used; nn>0; ) {
        struct modeNode *mNode = channel->members.list[--nn];
        if (IsService(mNode->user)) continue;
        KickChannelUser(mNode->user, channel, opserv, OSMSG_ILLEGAL_REASON);
    }
}

static void
opserv_channel_check(struct chanNode *newchan)
{
    unsigned int n;
    newchan->join_policer = policer_new(opserv_conf.join_policer_params, now);
    if ((newchan->bad_channel = opserv_bad_channel(newchan->name))) {
        if (newchan->name[0] == '#') {
            /* We can hack modes and ops and stuff, but we will cause
             * a HACK(2) if we join now (because it will send a CREATE
             * instead of a JOIN) .. so wait until the join check to
             * shut the channel down. */
        } else {
            for (n=newchan->members.used; n>0; ) {
                struct modeNode *mNode = newchan->members.list[--n];
                if (IsService(mNode->user)) continue;
                DelUser(mNode->user, opserv, true, OSMSG_ILLEGAL_REASON);
            }
        }
	return;
    }

    for (n=0; n<opserv_secret_words->used; n++) {
	if (strcasestr(newchan->name, opserv_secret_words->list[n])) {
	    newchan->modes |= MODE_SECRET;
	    irc_mode(self->numeric, newchan, "+s");
	}
    }
}

static void
opserv_del_channel(struct chanNode *chan)
{
    if (chan->join_policer) {
        policer_delete(chan->join_policer);
    } else {
        log(OS_LOG, LOG_ERROR, "No join policer found for channel %s\n", chan->name);
    }
}

static void
opserv_join_check(struct userNode *user, struct chanNode *channel, int is_burst)
{
    if (channel->bad_channel && !IsService(user)) {
        if (channel->name[0] == '#') {
            if (!GetUserMode(channel, opserv)) {
                opserv_shutdown_channel(channel);
            } else {
                opserv_notice(user, OSMSG_ILLEGAL_CHANNEL, channel->name);
                KickChannelUser(user, channel, opserv, OSMSG_ILLEGAL_REASON);
            }
        } else {
            DelUser(user, opserv, true, OSMSG_ILLEGAL_REASON);
        }
    }

    /* TODO: Figure out why the join_policer may be null here, and fix it */
    if (!is_burst && !policer_conforms(channel->join_policer, now, 1.0)) {
	if (!channel->join_flooded) {
	    channel->join_flooded = 1;
	    opserv_alert(OSMSG_JOIN_FLOOD, channel->name, channel->members.used);
	}
        log(OS_LOG, LOG_INFO, "Join to %s during flood: "IDENT_FORMAT"\n", channel->name, IDENT_DATA(user));
    } else {
        if (!channel->join_policer) {
            log(OS_LOG, LOG_ERROR, "No join policer found for channel %s\n", channel->name);
        }
	channel->join_flooded = 0;
    }
}


static OPSERV_FUNC(cmd_addbad)
{
    char *new_bad;
    unsigned int m, n;

    OPSERV_MIN_PARMS(2, false);

    for (n=1; n<argc; n++) {
	new_bad = strdup(argv[n]);
	/* loop over current bad words, see if we subsume any with this new one */
	for (m=0; m<opserv_bad_words->used; m++) {
	    if (strcasestr(opserv_bad_words->list[m], new_bad)) {
		string_list_delete(opserv_bad_words, m);
	    }
	}
	string_list_append(opserv_bad_words, new_bad);
    }
    opserv_notice(user, OSMSG_ADDED_BAD);
    return 1;
}

static OPSERV_FUNC(cmd_delbad)
{
    char *new_bad;
    unsigned int m, n;

    OPSERV_MIN_PARMS(2, false);

    for (n=1; n<argc; n++) {
	new_bad = argv[n];
	for (m=0; m<opserv_bad_words->used; m++) {
	    if (!strcmp(opserv_bad_words->list[m], new_bad)) {
		string_list_delete(opserv_bad_words, m);
	    }
	}
    }
    opserv_notice(user, OSMSG_REMOVED_BAD);
    return 1;
}

static OPSERV_FUNC(cmd_addsecret)
{
    char *new_secret;
    unsigned int m, n;

    OPSERV_MIN_PARMS(2, false);

    for (n=1; n<argc; n++) {
	new_secret = argv[n];
	/* loop over current secret words, see if we subsume any with this new one */
	for (m=0; m<opserv_secret_words->used; m++) {
	    if (strstr(opserv_secret_words->list[m], new_secret)) {
		string_list_delete(opserv_secret_words, m);
	    }
	}
	string_list_append(opserv_secret_words, new_secret);
    }
    opserv_notice(user, OSMSG_ADDED_SECRET);
    return 1;
}

static OPSERV_FUNC(cmd_delsecret)
{
    char *new_secret;
    unsigned int m, n;

    OPSERV_MIN_PARMS(2, false);

    for (n=1; n<argc; n++) {
	new_secret = argv[n];
	for (m=0; m<opserv_secret_words->used; m++) {
	    if (!strcmp(opserv_secret_words->list[m], new_secret)) {
		string_list_delete(opserv_secret_words, m);
	    }
	}
    }
    opserv_notice(user, OSMSG_REMOVED_SECRET);
    return 1;
}

static void
opserv_expire_trusted_host(void *data)
{
    struct trusted_host *th = data;
    dict_remove(opserv_trusted_hosts, th->hostname);
}

static void
opserv_add_trusted_host(const char *hostname, unsigned int limit, time_t expires, const char *reason)
{
    struct trusted_host *th;
    th = malloc(sizeof(*th));
    if (!th) return;
    th->hostname = strdup(hostname);
    th->limit = limit;
    th->expires = expires;
    safestrncpy(th->reason, reason, sizeof(th->reason));
    dict_insert(opserv_trusted_hosts, th->hostname, th);
    if (th->expires) timeq_add(th->expires, opserv_expire_trusted_host, th);
}

static OPSERV_FUNC(cmd_addtrust)
{
    OPSERV_MIN_PARMS(5, false);

    if (dict_find(opserv_trusted_hosts, argv[1], NULL)) {
        return opserv_notice(user, OSMSG_ALREADY_TRUSTED, argv[1]);
    }
    unsplit_string(argv+4, argc-4);
    opserv_add_trusted_host(argv[1], strtoul(argv[2], NULL, 0), now + ParseInterval(argv[3]), argv[4]);
    return opserv_notice(user, OSMSG_ADDED_TRUSTED);
}

static OPSERV_FUNC(cmd_deltrust)
{
    unsigned int n;

    OPSERV_MIN_PARMS(2, false);

    for (n=1; n<argc; n++) {
        struct trusted_host *th = dict_find(opserv_trusted_hosts, argv[n], NULL);
        if (!th) continue;
        if (th->expires) timeq_del(th->expires, opserv_expire_trusted_host, th, 0);
	dict_remove(opserv_trusted_hosts, argv[n]);
    }
    return opserv_notice(user, OSMSG_REMOVED_TRUSTED);
}

/* This doesn't use dict_t because it's a little simpler to open-code the
 * comparisons (and simpler arg-passing for the ADD subcommand).
 */
static OPSERV_FUNC(cmd_clone)
{
    int i;
    struct userNode *clone;

    OPSERV_MIN_PARMS(3, false);

    clone = GetUserH(argv[2]);
    if (!strcasecmp(argv[1], "ADD")) {
	if (argc < 5) {
	    opserv_notice(user, MSG_MISSING_PARAMS, argv[1]);
	    OPSERV_SYNTAX();
	    return 0;
	}
	if (clone) {
	    opserv_notice(user, OSMSG_CLONE_EXISTS, argv[2]);
	    return 0;
	}
	unsplit_string(argv+4, argc-4);
	for (i=0; argv[3][i]; i++) {
	    if (argv[3][i] == '@') {
		argv[3][i] = 0;
		i++;
		break;
	    }
	}
	if (!argv[3][i]) {
	    opserv_notice(user, OSMSG_NOT_A_HOSTMASK);
	    return 0;
	}
	clone = AddClone(argv[2], argv[3], argv[3]+i, argv[4]);
	opserv_notice(user, OSMSG_CLONE_ADDED, clone->nick);
	return 1;
    }
    if (!clone) {
	opserv_notice(user, MSG_NICK_UNKNOWN, argv[2]);
	return 0;
    }
    if (clone->uplink != self || IsService(clone)) {
	opserv_notice(user, OSMSG_NOT_A_CLONE, clone->nick);
	return 0;
    }
    if (!strcasecmp(argv[1], "REMOVE")) {
	const char *reason;
	if (argc > 3) {
	    unsplit_string(argv+3, argc-3);
	    reason = argv[3];
	} else {
	    char *tmp;
	    tmp = alloca(strlen(clone->nick) + strlen(OSMSG_PART_REASON));
	    sprintf(tmp, OSMSG_PART_REASON, clone->nick);
	    reason = tmp;
	}
	DelUser(clone, NULL, true, reason);
	opserv_notice(user, OSMSG_CLONE_REMOVED, argv[2]);
	return 1;
    }
    if (argc < 4) {
	opserv_notice(user, MSG_MISSING_PARAMS, argv[1]);
	OPSERV_SYNTAX();
	return 0;
    }
    channel = GetChannel(argv[3]);
    if (!strcasecmp(argv[1], "JOIN")) {
	if (!channel
	    && !(channel = AddChannel(argv[3], now, "", "", 0, "", NULL))) {
	    opserv_notice(user, MSG_CHANNEL_UNKNOWN, argv[3]);
	    return 0;
	}
	AddChannelUser(clone, channel, 0);
	opserv_notice(user, OSMSG_CLONE_JOINED, clone->nick, channel->name);
	return 1;
    }
    if (!strcasecmp(argv[1], "PART")) {
	if (!channel) {
	    opserv_notice(user, MSG_CHANNEL_UNKNOWN, argv[3]);
	    return 0;
	}
	if (!GetUserMode(channel, clone)) {
	    opserv_notice(user, OSMSG_NOT_ON_CHANNEL, clone->nick, channel->name);
	    return 0;
	}
	DelChannelUser(clone, channel, "Leaving.", 0);
	opserv_notice(user, OSMSG_CLONE_PARTED, clone->nick, channel->name);
	return 1;
    }
    if (!strcasecmp(argv[1], "OP")) {
	if (!channel) {
	    opserv_notice(user, MSG_CHANNEL_UNKNOWN, argv[3]);
	    return 0;
	}
	if (!GetUserMode(channel, clone)) {
	    opserv_notice(user, OSMSG_NOT_ON_CHANNEL, clone->nick, channel->name);
	    return 0;
	}
	AddChannelOp(1, &clone, channel, oper_numeric(channel), true);
	opserv_notice(user, OSMSG_OPS_GIVEN, channel->name, clone->nick);
	return 1;
    }
    if (argc < 5) {
	opserv_notice(user, MSG_MISSING_PARAMS, argv[1]);
	OPSERV_SYNTAX();
	return 0;
    }
    if (!strcasecmp(argv[1], "SAY")) {
	unsplit_string(argv+4, argc-4);
	irc_privmsg(clone->numeric, argv[3], argv[4]);
	opserv_notice(user, OSMSG_CLONE_SAID, clone->nick, argv[3]);
	return 1;
    }
    opserv_notice(user, OSMSG_UNKNOWN_SUBCOMMAND, argv[1], argv[0]);
    return 0;
}

static OPSERV_FUNC(cmd_version)
{
    extern char *GMSG_VERSION_ID;
    extern char *NSMSG_VERSION_ID;
    extern char *CSMSG_VERSION_ID;

    (void)channel; (void)argc; (void)argv;
    opserv_notice(user, MSG_VERSION);
    opserv_notice(user, GMSG_VERSION_ID);
    opserv_notice(user, NSMSG_VERSION_ID);
    opserv_notice(user, OSMSG_VERSION_ID);
    opserv_notice(user, CSMSG_VERSION_ID);

    return 1;
}

static int
append_entry(const char *cmdname, void *data, void *extra)
{
    struct opserv_cmd *cmd = data;
    struct string_buffer *sbuf = extra;
    char buffer[80];
    int pos;
    if (!cmd->func) return 0;
    pos = snprintf(buffer, sizeof(buffer), OSMSG_INDEX_FORMAT, cmdname, cmd->min_level, (cmd->disabled?'*':' '));
    while (pos<16) buffer[pos++] = ' ';
    buffer[pos] = 0;
    string_buffer_append_string(sbuf, buffer);
    return 0;
}

static char*
opserv_help_expand(char *variable)
{
    if (!strcasecmp(variable, "index")) {
        struct string_buffer sbuf;
        string_buffer_init(&sbuf);
        string_buffer_append_string(&sbuf, "Commands:\n");
        dict_foreach(opserv_func_dict, append_entry, &sbuf);
        return sbuf.list;
    }
    return NULL;
}

static void
opserv_helpfile_read(void)
{
    opserv_helpfile = open_helpfile(OPSERV_HELPFILE_NAME, opserv_help_expand);
}

static OPSERV_FUNC(cmd_help)
{
    const char *topic;

    (void)channel;
    if (argc < 2) {
	topic = NULL;
    } else {
	topic = argv[1];
	unsplit_string(argv+1, argc-1);
    }
    opserv_help(user, topic, 1);
    return 1;
}

static OPSERV_FUNC(cmd_readhelp)
{
    struct timeval start, stop;
    struct helpfile *old_helpfile = opserv_helpfile;
    (void)channel; (void)argc; (void)argv; /* ignore them!  ignore them all! */
    gettimeofday(&start, NULL);
    opserv_helpfile_read();
    if (opserv_helpfile) {
	close_helpfile(old_helpfile);
    } else {
	opserv_helpfile = old_helpfile;
    }
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
	stop.tv_sec -= 1;
	stop.tv_usec += 1000000;
    }
    opserv_notice(user, OSMSG_READHELP_SUCCESS, stop.tv_sec, stop.tv_usec/1000);
    return 1;
}

int opserv_command_search(const char *command)
{
    struct opserv_cmd *cmd = dict_find(opserv_func_dict, command, NULL);
    return cmd ? 1 : 0;
}

int opserv_add_alias(const char *alias, const char *command)
{
    struct opserv_cmd *cmd = dict_find(opserv_func_dict, command, NULL);
    if(!cmd) return 0;

    dict_insert(opserv_func_dict, alias, cmd);
    return 1;
}

int opserv_del_alias(const char *alias)
{
    return dict_remove(opserv_func_dict, alias);
}

struct opserv_cmd *
opserv_define_func(const char *name, opserv_func_t *func, int min_level)
{
    struct opserv_cmd *cmd = malloc(sizeof(struct opserv_cmd));
    cmd->min_level = min_level;
    cmd->disabled = 0;
    cmd->weight = 1.0;
    cmd->func = func;
    dict_insert(opserv_func_dict, name, cmd);
    return cmd;
}

int
set_cmd_options(const char *key, void *data, void *extra)
{
    struct record_data *rd = data;
    struct opserv_cmd *cmd = dict_find(extra, key, NULL);
    char *str;
    if (!cmd) {
	log(OS_LOG, LOG_ERROR, "OpServ DB had command entry for unknown command `%s'\n", key);
	return 0;
    }
    if ((str = database_get_data(rd->d.object, KEY_MIN_LEVEL, RECDB_QSTRING))) {
	cmd->min_level = strtoul(str, NULL, 0);
    }
    if ((str = database_get_data(rd->d.object, KEY_DISABLED, RECDB_QSTRING))) {
	cmd->disabled = strtoul(str, NULL, 0);
    }
    if ((str = database_get_data(rd->d.object, KEY_WEIGHT, RECDB_QSTRING))) {
	cmd->weight = strtod(str, NULL);
    }
    return 0;
}

int add_reserved(const char *key, void *data, void *extra)
{
    struct record_data *rd = data;
    const char *username, *hostname, *desc;
    struct userNode *reserve;
    username = database_get_data(rd->d.object, KEY_USERNAME, RECDB_QSTRING);
    if (!username) {
	log(OS_LOG, LOG_ERROR, "missing username for reserve of %s", key);
	return 0;
    }
    hostname = database_get_data(rd->d.object, KEY_HOSTNAME, RECDB_QSTRING);
    if (!hostname) {
	log(OS_LOG, LOG_ERROR, "missing hostname for reserve of %s", key);
	return 0;
    }
    desc = database_get_data(rd->d.object, KEY_DESC, RECDB_QSTRING);
    if (!desc) {
	log(OS_LOG, LOG_ERROR, "missing description for reserve of %s", key);
	return 0;
    }
    reserve = AddClone(key, username, hostname, desc);
    if (database_get_data(rd->d.object, KEY_PERSISTENT, RECDB_QSTRING)) {
	reserve->modes |= FLAGS_PERSISTENT;
    }
    dict_insert(extra, reserve->nick, reserve);
    return 0;
}

static int ungag_helper_func(struct userNode *match, void *extra);

static unsigned int
gag_free(struct gag_entry *gag)
{
    struct gag_entry *prev;
    unsigned int ungagged;
    discrim_t discrim;
    char *str;

    /* Find previous gag (if there is one) */
    if (gagList == gag) {
        prev = NULL;
    } else {
        for (prev = gagList; prev->next != gag; prev = prev->next) ;
    }

    /* Remove from gagList */
    if (prev) {
	prev->next = gag->next;
    } else {
	gagList = gag->next;
    }

    /* Ungag anybody affected by it. */
    discrim = calloc(1, sizeof(*discrim));
    discrim->limit = dict_size(clients);
    str = strdup(gag->mask);
    if (split_ircmask(str, &discrim->mask_nick, &discrim->mask_ident, &discrim->mask_host)) {
        if (discrim->mask_host && !discrim->mask_host[strspn(discrim->mask_host, "0123456789.?*")]) {
            if (!parse_ipmask(discrim->mask_host, &discrim->ip_addr, &discrim->ip_mask)) {
                free(discrim);
                free(str);
                return 0;
            }
            discrim->mask_host = 0;
        }
        ungagged = opserv_discrim_search(discrim, ungag_helper_func, NULL);
    } else {
	log(OS_LOG, LOG_ERROR, "Couldn't split IRC mask for gag %s!\n", gag->mask);
        ungagged = 0;
    }

    /* Deallocate storage */
    free(gag->reason);
    free(gag->owner);
    free(gag->mask);
    free(gag);

    return ungagged;
}

static void
gag_expire(void *data)
{
    gag_free(data);
}

static unsigned int
gag_create(const char *mask, const char *owner, const char *reason, time_t expires)
{
    struct gag_entry *gag;

    /* Create gag and put it into linked list */
    gag = calloc(1, sizeof(*gag));
    gag->mask = strdup(mask);
    gag->owner = strdup(owner ? owner : "<unknown>");
    gag->reason = strdup(reason ? reason : "<unknown>");
    gag->expires = expires;
    if (gag->expires) timeq_add(gag->expires, gag_expire, gag);
    gag->next = gagList;
    gagList = gag;
    /* If we're linked, see if who the gag applies to */
    if (self->uplink) {
        discrim_t discrim;
        char *maskdup;
        unsigned int gagged;

        discrim = calloc(1, sizeof(*discrim));
        discrim->limit = dict_size(clients);
        maskdup = strdup(mask);
        split_ircmask(maskdup, &discrim->mask_nick, &discrim->mask_ident, &discrim->mask_host);
        if (discrim->mask_host && !discrim->mask_host[strspn(discrim->mask_host, "0123456789.?*")]) {
            if (!parse_ipmask(discrim->mask_host, &discrim->ip_addr, &discrim->ip_mask)) {
                free(discrim);
                free(maskdup);
                return 0;
            }
            discrim->mask_host = 0;
        }
        gagged = opserv_discrim_search(discrim, gag_helper_func, gag);
        free(discrim);
        free(maskdup);
        return gagged;
    } else {
        return 0;
    }
}

static int
add_gag_helper(const char *key, void *data, void *extra)
{
    struct record_data *rd = data;
    char *owner, *reason, *expstr;
    time_t expires;

    (void)extra;
    owner = database_get_data(rd->d.object, KEY_OWNER, RECDB_QSTRING);
    reason = database_get_data(rd->d.object, KEY_REASON, RECDB_QSTRING);
    expstr = database_get_data(rd->d.object, KEY_EXPIRES, RECDB_QSTRING);
    expires = expstr ? strtoul(expstr, NULL, 0) : 0;
    gag_create(key, owner, reason, expires);

    return 0;
}

static struct opserv_user_alert *
opserv_add_user_alert(struct userNode *req, const char *name, opserv_alert_reaction reaction, const char *text_discrim)
{
    unsigned int wordc;
    unsigned char *wordv[MAXNUMPARAMS], *discrim_copy;
    struct opserv_user_alert *alert;
    char *name_dup;

    if (dict_find(opserv_user_alerts, name, NULL)) {
	opserv_notice(req, OSMSG_ALERT_EXISTS, name);
        return NULL;
    }
    alert = malloc(sizeof(*alert));
    alert->owner = strdup(req->handle_info ? req->handle_info->handle : req->nick);
    alert->text_discrim = strdup(text_discrim);
    discrim_copy = strdup(text_discrim); /* save a copy of the discrim */
    wordc = split_line(discrim_copy, false, ArrayLength(wordv), wordv);
    alert->discrim = opserv_discrim_create(req, wordc, wordv);
    if (!alert->discrim) {
        free(alert->text_discrim);
        free(discrim_copy);
        free(alert);
        return NULL;;
    }
    name_dup = strdup(name);
    if (!alert->discrim->reason) alert->discrim->reason = name_dup;
    alert->reaction = reaction;
    dict_insert(opserv_user_alerts, name_dup, alert);
    if (alert->discrim->mask_nick) dict_insert(opserv_nick_based_alerts, name_dup, alert);
    return alert;
}

static int
add_chan_warn(const char *key, void *data, void *extra)
{
    struct record_data *rd = data;
    char *reason = GET_RECORD_QSTRING(rd);

    (void)extra;

    /* i hope this cant happen */
    if (!reason) reason = strdup("No Reason");

    dict_insert(opserv_chan_warn, strdup(key), strdup(reason));
    return 0;
}

static int
add_user_alert(const char *key, void *data, void *extra)
{
    dict_t alert_dict;
    const char *discrim, *react, *owner;
    opserv_alert_reaction reaction;
    struct opserv_user_alert *alert;

    (void)extra;
    /* TODO: gripe about db error */
    if (!(alert_dict = GET_RECORD_OBJECT((struct record_data *)data))) return 1;
    discrim = database_get_data(alert_dict, KEY_DISCRIM, RECDB_QSTRING);
    react = database_get_data(alert_dict, KEY_REACTION, RECDB_QSTRING);
    if (!react || !strcasecmp(react, "notice")) {
        reaction = REACT_NOTICE;
    } else if (!strcasecmp(react, "gline")) {
        reaction = REACT_GLINE;
    } else {
        log(OS_LOG, LOG_ERROR, "Invalid reaction %s for alert %s\n", react, key);
        return 0;
    }
    alert = opserv_add_user_alert(opserv, key, reaction, discrim);
    owner = database_get_data(alert_dict, KEY_OWNER, RECDB_QSTRING);
    free(alert->owner);
    alert->owner = strdup(owner ? owner : "<unknown>");
    return 0;
}

static int
trusted_host_read(const char *host, void *data, void *extra)
{
    struct record_data *rd = data;
    const char *limit, *expstr, *reason;
    time_t expires;

    (void)extra;
    if (rd->type == RECDB_QSTRING) {
        /* old style host by itself */
        limit = GET_RECORD_QSTRING(rd);
        expires = 0;
        reason = "<unknown>";
    } else if (rd->type == RECDB_OBJECT) {
        dict_t obj = GET_RECORD_OBJECT(rd);
        /* new style structure */
        limit = database_get_data(obj, KEY_LIMIT, RECDB_QSTRING);
        expstr = database_get_data(obj, KEY_EXPIRES, RECDB_QSTRING);
        expires = expstr ? ParseInterval(expstr) : 0;
        reason = database_get_data(obj, KEY_REASON, RECDB_QSTRING);
    } else return 0;

    if (!limit || (expires && (expires < now))) return 0;
    opserv_add_trusted_host(host, strtoul(limit, NULL, 0), expires, reason);
    return 0;
}

static void
opserv_db_read(void)
{
    dict_t conf_db, object;
    struct string_list *dbsl;
    conf_db = parse_database(OPSERV_DB_NAME);
    opserv_trusted_hosts = dict_new();
    dict_set_free_keys(opserv_trusted_hosts, free);
    dict_set_free_data(opserv_trusted_hosts, free);
    opserv_chan_warn = dict_new();
    dict_set_free_keys(opserv_chan_warn, free);
    dict_set_free_data(opserv_chan_warn, free);
    opserv_nick_based_alerts = dict_new();
    opserv_user_alerts = dict_new();
    dict_set_free_keys(opserv_user_alerts, free);
    dict_set_free_data(opserv_user_alerts, opserv_free_user_alert);
    if (conf_db) {
	if ((object = database_get_data(conf_db, KEY_COMMANDS, RECDB_OBJECT))) {
	    dict_foreach(object, set_cmd_options, opserv_func_dict);
	}
	if ((object = database_get_data(conf_db, KEY_RESERVES, RECDB_OBJECT))) {
	    dict_foreach(object, add_reserved, opserv_reserved_nick_dict);
	}
	if ((dbsl = database_get_data(conf_db, KEY_BAD_WORDS, RECDB_STRING_LIST))) {
	    opserv_bad_words = string_list_copy(dbsl);
	} else {
	    opserv_bad_words = alloc_string_list(8);
	}
       if ((object = database_get_data(conf_db, KEY_MAX_CLIENTS, RECDB_OBJECT))) {
           char *str;
           if ((str = database_get_data(object, KEY_MAX, RECDB_QSTRING))) max_clients = atoi(str);
           if ((str = database_get_data(object, KEY_TIME, RECDB_QSTRING))) max_clients_time = atoi(str);
       }
	if ((dbsl = database_get_data(conf_db, KEY_SECRET_WORDS, RECDB_STRING_LIST))) {
	    opserv_secret_words = string_list_copy(dbsl);
	} else {
	    opserv_secret_words = alloc_string_list(8);
	}
	if ((object = database_get_data(conf_db, KEY_TRUSTED_HOSTS, RECDB_OBJECT))) {
	    dict_foreach(object, trusted_host_read, opserv_trusted_hosts);
	}
	if ((object = database_get_data(conf_db, KEY_GAGS, RECDB_OBJECT))) {
	    dict_foreach(object, add_gag_helper, NULL);
	}
	if ((object = database_get_data(conf_db, KEY_ALERTS, RECDB_OBJECT))) {
	    dict_foreach(object, add_user_alert, NULL);
	}
	if ((object = database_get_data(conf_db, KEY_WARN, RECDB_OBJECT))) {
	    dict_foreach(object, add_chan_warn, NULL);
	}
	free_database(conf_db);
    } else {
	opserv_bad_words = alloc_string_list(8);
	opserv_secret_words = alloc_string_list(8);
    }
}

static int
insert_func_entry(const char *cmd_name, void *data, void *extra)
{
    struct opserv_cmd *cmd = data;
    char buff[32];
    dict_t cmd_db = alloc_object();
    snprintf(buff, sizeof(buff), "%d", cmd->min_level);
    dict_insert(cmd_db, KEY_MIN_LEVEL, alloc_record_data_qstring(buff));
    dict_insert(cmd_db, KEY_DISABLED, alloc_record_data_qstring(cmd->disabled?"1":"0"));
    snprintf(buff, sizeof(buff), "%g", cmd->weight);
    dict_insert(cmd_db, KEY_WEIGHT, alloc_record_data_qstring(buff));
    dict_insert(extra, cmd_name, alloc_record_data_object(cmd_db));
    return 0;
}

static int
insert_reserve_entry(const char *nick, void *data, void *extra)
{
    struct userNode *user = data;
    dict_t node = alloc_object();
    dict_insert(node, KEY_USERNAME, alloc_record_data_qstring(user->username));
    dict_insert(node, KEY_HOSTNAME, alloc_record_data_qstring(user->hostname));
    dict_insert(node, KEY_DESC, alloc_record_data_qstring(user->info));
    if (user->modes & FLAGS_PERSISTENT) {
	dict_insert(node, KEY_PERSISTENT, alloc_record_data_qstring("yes"));
    }
    dict_insert(extra, nick, alloc_record_data_object(node));
    return 0;
}

static void
gag_write(dict_t gags)
{
    struct gag_entry *gag;

    for (gag = gagList; gag; gag = gag->next) {
	dict_t node = alloc_object();
	dict_insert(node, KEY_OWNER, alloc_record_data_qstring(gag->owner));
	dict_insert(node, KEY_REASON, alloc_record_data_qstring(gag->reason));
        if (gag->expires) {
            char buf[16];
            snprintf(buf, sizeof(buf), FMT_TIME_T, gag->expires);
            dict_insert(node, KEY_EXPIRES, alloc_record_data_qstring(buf));
        }
	dict_insert(gags, gag->mask, alloc_record_data_object(node));
    }
}

static int
trusted_host_write(const char *host, void *data, void *extra)
{
    struct trusted_host *th = data;
    dict_t subobj = alloc_object();
    char number[16];

    snprintf(number, sizeof(number), "%lu", th->limit);
    dict_insert(subobj, KEY_LIMIT, alloc_record_data_qstring(number));
    snprintf(number, sizeof(number), FMT_TIME_T, th->expires);
    dict_insert(subobj, KEY_EXPIRES, alloc_record_data_qstring(number));
    dict_insert(subobj, KEY_REASON, alloc_record_data_qstring(th->reason));
    dict_insert(extra, host, alloc_record_data_object(subobj));
    return 0;
}

static int
chan_warn_write(const char *name, void *data, void *extra)
{
    char *reason = data;
    dict_insert(extra, name, alloc_record_data_qstring(reason));
    return 0;
}

static int
user_alert_write(const char *name, void *data, void *extra)
{
    struct opserv_user_alert *alert = data;
    const char *reaction;
    dict_t object = alloc_object();
    dict_insert(object, KEY_DISCRIM, alloc_record_data_qstring(alert->text_discrim));
    dict_insert(object, KEY_OWNER, alloc_record_data_qstring(alert->owner));
    switch (alert->reaction) {
    case REACT_NOTICE: reaction = "notice"; break;
    case REACT_GLINE: reaction = "gline"; break;
    default:
        reaction = NULL;
        log(OS_LOG, LOG_ERROR, "Invalid reaction type %d for alert %s\n", alert->reaction, name);
        break;
    }
    if (reaction) dict_insert(object, KEY_REACTION, alloc_record_data_qstring(reaction));
    dict_insert(extra, name, alloc_record_data_object(object));
    return 0;
}

static void
max_clients_write(void *extra)
{
    char buff[16];
    snprintf(buff, sizeof(buff), "%d", max_clients);   
    dict_insert(extra, KEY_MAX, alloc_record_data_qstring(buff));
    snprintf(buff, sizeof(buff), FMT_TIME_T, max_clients_time);
    dict_insert(extra, KEY_TIME, alloc_record_data_qstring(buff));
}

static void
opserv_db_write(void)
{
    FILE *f;
    dict_t db = alloc_database(), subdb;
    /* write out the commands */
    subdb = alloc_object();
    dict_foreach(opserv_func_dict, insert_func_entry, subdb);
    dict_insert(db, KEY_COMMANDS, alloc_record_data_object(subdb));
    /* write out any reserved nicks */
    subdb = alloc_object();
    dict_foreach(opserv_reserved_nick_dict, insert_reserve_entry, subdb);
    dict_insert(db, KEY_RESERVES, alloc_record_data_object(subdb));
    /* bad and secret word lists */
    dict_insert(db, KEY_BAD_WORDS, alloc_record_data_string_list(string_list_copy(opserv_bad_words)));
    dict_insert(db, KEY_SECRET_WORDS, alloc_record_data_string_list(string_list_copy(opserv_secret_words)));
    /* trusted hosts takes a little more work */
    subdb = alloc_object();
    dict_foreach(opserv_trusted_hosts, trusted_host_write, subdb);
    dict_insert(db, KEY_TRUSTED_HOSTS, alloc_record_data_object(subdb));
    /* gags */
    subdb = alloc_object();
    gag_write(subdb);
    dict_insert(db, KEY_GAGS, alloc_record_data_object(subdb));
    /* channel warnings */
    subdb = alloc_object();
    dict_foreach(opserv_chan_warn, chan_warn_write, subdb);
    dict_insert(db, KEY_WARN, alloc_record_data_object(subdb));
    /* alerts */
    subdb = alloc_object();
    dict_foreach(opserv_user_alerts, user_alert_write, subdb);
    dict_insert(db, KEY_ALERTS, alloc_record_data_object(subdb));
    /* max clients */
    subdb = alloc_object(); 
    max_clients_write(subdb);
    dict_insert(db, KEY_MAX_CLIENTS, alloc_record_data_object(subdb));
    /* now write the entire database */
    f = fopen(TEMP_DB_NAME, "w");
    write_database(f, db);
    fclose(f);
    free_database(db);
    if (rename(TEMP_DB_NAME, OPSERV_DB_NAME) == -1) {
	log(NS_LOG, LOG_ERROR, "unable to rename new db to old: %s\n", strerror(errno));
    }
    last_db_write = now;
}

void
opserv_timed_write(void *data)
{
    (void)data;

    opserv_db_write();
    log(OS_LOG, LOG_INFO, "Successfully made scheduled backup of OpServ database.\n");
    timeq_add(now + opserv_conf.db_backup_frequency, opserv_timed_write, NULL);
}

static OPSERV_FUNC(cmd_write)
{
    struct timeval start, stop;
    (void)argc; (void)argv; (void)channel;
    gettimeofday(&start, NULL);
    opserv_db_write();
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
	stop.tv_sec -= 1;
	stop.tv_usec += 1000000;
    }
    opserv_notice(user, OSMSG_WRITE_SUCCESS, stop.tv_sec, stop.tv_usec/1000);
    return 1;
}

static OPSERV_FUNC(cmd_writeall)
{
    struct timeval start, stop;
    (void)argc; (void)argv; (void)channel;
    gettimeofday(&start, NULL);
    write_databases();
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
	stop.tv_sec -= 1;
	stop.tv_usec += 1000000;
    }
    opserv_notice(user, OSMSG_WRITEALL_SUCCESS, stop.tv_sec, stop.tv_usec/1000);
    return 1;
}

static int
query_keys_helper(const char *key, void *data, void *extra)
{
    (void)data;
    opserv_notice(extra, OSMSG_OPTION_LIST_ENTRY, key);
    return 0;
}

static OPSERV_FUNC(cmd_query)
{
    struct record_data *rd;
    unsigned int i;

    (void)channel;

    if(argc < 2)
    {
	opserv_notice(user, OSMSG_OPTION_ROOT);
	conf_enum_root(query_keys_helper, user);
	return 1;
    }

    unsplit_string(argv+1, argc-1);
    if(!(rd = conf_get_node(argv[1])))
    {
	opserv_notice(user, OSMSG_UNKNOWN_OPTION, argv[1]);
	return 0;
    }

    if(rd->type == RECDB_QSTRING)
    {
	opserv_notice(user, OSMSG_OPTION_IS, argv[1], rd->d.qstring);
    }
    else if(rd->type == RECDB_STRING_LIST)
    {
	opserv_notice(user, OSMSG_OPTION_LIST, argv[1]);

	if (rd->d.slist->used)
	{
	    for (i=0; i<rd->d.slist->used; i++)
	    {
		opserv_notice(user, OSMSG_OPTION_LIST_ENTRY, rd->d.slist->list[i]);
	    }
	}
	else
	{
	    opserv_notice(user, "(Empty list.)");
	}
    }
    else if(rd->type == RECDB_OBJECT)
    {
	opserv_notice(user, OSMSG_OPTION_KEYS, argv[1]);
	dict_foreach(rd->d.object, query_keys_helper, user);
    }

    return 1;
}

static OPSERV_FUNC(cmd_set)
{
    struct record_data *rd;

    OPSERV_MIN_PARMS(3, false);

    /* I originally wanted to be able to fully manipulate the config
       db with this, but i wussed out. feel free to fix this - you'll
       need to handle quoted strings which have been split, and likely
       invent a syntax for it. -Zoot */

    if(!(rd = conf_get_node(argv[1])))
    {
	opserv_notice(user, OSMSG_SET_NOT_SET, argv[1]);
	return 0;
    }

    if(rd->type != RECDB_QSTRING)
    {
	opserv_notice(user, OSMSG_SET_BAD_TYPE, argv[1]);
	return 0;
    }

    free(rd->d.qstring);
    rd->d.qstring = strdup(argv[2]);
    conf_call_reload_funcs();
    opserv_notice(user, OSMSG_SET_SUCCESS, argv[1], argv[2]);

    return 1;
}

static discrim_t
opserv_discrim_create(struct userNode *user, unsigned int argc, unsigned char *argv[])
{
    unsigned int i, j;
    discrim_t discrim;

    discrim = calloc(1, sizeof(*discrim));
    discrim->limit = 250;
    discrim->max_level = -1;

    for (i=0; i<argc; i++) {
        /* Assume all criteria require arguments. */
        if (i == argc - 1) {
            opserv_notice(user, MSG_MISSING_PARAMS, argv[i]);
            return NULL;
        }
	if (!strcasecmp(argv[i], "mask")) {
	    if (!is_ircmask(argv[++i])) {
		opserv_notice(user, OSMSG_INVALID_IRCMASK, argv[i]);
		goto fail;
	    }

	    if (!split_ircmask(argv[i],
                               &discrim->mask_nick,
                               &discrim->mask_ident,
                               &discrim->mask_host)) {
		opserv_notice(user, OSMSG_INVALID_IRCMASK, argv[i]);
		goto fail;
	    }
	} else if (strcasecmp(argv[i], "nick") == 0) {
	    discrim->mask_nick = argv[++i];
	} else if (strcasecmp(argv[i], "ident") == 0) {
	    discrim->mask_ident = argv[++i];
	} else if (strcasecmp(argv[i], "host") == 0) {
	    discrim->mask_host = argv[++i];
	} else if (strcasecmp(argv[i], "info") == 0) {
	    discrim->mask_info = argv[++i];
	} else if (strcasecmp(argv[i], "server") == 0) {
	    discrim->server = argv[++i];
	} else if (strcasecmp(argv[i], "ip") == 0) {
	    j = parse_ipmask(argv[++i], &discrim->ip_addr, &discrim->ip_mask);
	    if (!j) {
		opserv_notice(user, OSMSG_NOT_IP);
		goto fail;
	    }
	} else if (strcasecmp(argv[i], "channel") == 0) {
	    discrim->channel = GetChannel(argv[++i]);
	    if (!discrim->channel) {
		opserv_notice(user, MSG_CHANNEL_UNKNOWN, argv[i]);
		goto fail;
	    }
	} else if (strcasecmp(argv[i], "limit") == 0) {
	    discrim->limit = strtoul(argv[++i], NULL, 10);
	} else if (strcasecmp(argv[i], "last") == 0) {
	    discrim->last = now - ParseInterval(argv[++i]);
        } else if (strcasecmp(argv[i], "reason") == 0) {
            unsplit_string(argv+i+1, argc-i-1);
            discrim->reason = argv[++i];
            i = argc;
        } else if (strcasecmp(argv[i], "access") == 0) {
            const unsigned char *cmp = argv[++i];
            if (cmp[0] == '<') {
                if (discrim->min_level == 0) discrim->min_level = 1;
                if (cmp[1] == '=') {
                    discrim->max_level = strtoul(cmp+2, NULL, 0);
                } else {
                    discrim->max_level = strtoul(cmp+1, NULL, 0) - 1;
                }
            } else if (cmp[0] == '=') {
                discrim->min_level = discrim->max_level = strtoul(cmp+1, NULL, 0);
            } else if (cmp[0] == '>') {
                if (cmp[1] == '=') {
                    discrim->min_level = strtoul(cmp+2, NULL, 0);
                } else {
                    discrim->min_level = strtoul(cmp+1, NULL, 0) + 1;
                }
            } else {
                opserv_notice(user, MSG_INVALID_CRITERIA, cmp);
            }
        } else if ((strcasecmp(argv[i], "abuse") == 0)
                   && (strcasecmp(argv[++i], "opers") == 0)) {
            discrim->match_opers = 1;
	} else {
	    opserv_notice(user, MSG_INVALID_CRITERIA, argv[i]);
	    goto fail;
	}
    }

    if (discrim->mask_nick && !strcmp(discrim->mask_nick, "*")) {
	discrim->mask_nick = 0;
    }
    if (discrim->mask_ident && !strcmp(discrim->mask_ident, "*")) {
        discrim->mask_ident = 0;
    }
    if (discrim->mask_info && !strcmp(discrim->mask_info, "*")) {
	discrim->mask_info = 0;
    }
    if (discrim->mask_host && !discrim->mask_host[strspn(discrim->mask_host, "*.")]) {
        discrim->mask_host = 0;
    }
    return discrim;
  fail:
    free(discrim);
    return NULL;
}

static int
discrim_match(discrim_t discrim, struct userNode *user)
{
    unsigned int access;
    if ((discrim->last && (discrim->last >= user->timestamp))
	|| (discrim->mask_nick && !match_ircglob(user->nick, discrim->mask_nick))
	|| (discrim->mask_ident && !match_ircglob(user->username, discrim->mask_ident))
	|| (discrim->mask_host && !match_ircglob(user->hostname, discrim->mask_host))
	|| (discrim->mask_info && !match_ircglob(user->info, discrim->mask_info))
	|| (discrim->server && !match_ircglob(user->uplink->name, discrim->server))
	|| (discrim->ip_mask && !MATCH_IPMASK(user->ip, discrim->ip_addr, discrim->ip_mask))) {
	return 0;
    }
    access = user->handle_info ? user->handle_info->opserv_level : 0;
    if ((access < discrim->min_level)
        || (access > discrim->max_level)) {
        return 0;
    }
    return 1;
}

struct discrim_apply_info {
    discrim_t discrim;
    struct userList *matched;
};

static int
foreach_discrim_helper(const char *nick, void *vuser, void *extra)
{
    struct userNode *user = vuser;
    struct discrim_apply_info *dai = extra;
    (void)nick;
    if (discrim_match(dai->discrim, user)) {
        userList_append(dai->matched, user);
    }
    return dai->matched->used >= dai->discrim->limit;
}

static unsigned int
opserv_discrim_search(discrim_t discrim, discrim_search_func dsf, void *data)
{
    unsigned int n, count;
    struct userList matched;

    userList_init(&matched);
    /* Try most optimized search methods first */
    if (discrim->channel) {
	unsigned long n;
	for (n=0;
	     (n < discrim->channel->members.used)
                 && (matched.used < discrim->limit);
	     n++) {
	    struct modeNode *mn = discrim->channel->members.list[n];
	    if (discrim_match(discrim, mn->user)) {
                userList_append(&matched, mn->user);
	    }
	}
    } else {
        struct discrim_apply_info dai;
        dai.discrim = discrim;
        dai.matched = &matched;
	dict_foreach(clients, foreach_discrim_helper, &dai);
    }

    for (n=0; n<matched.used; n++) {
        if(dsf(matched.list[n], data)) {
	    /* If a search function returns true, it ran into a
	       problem. Stop going through the list. */
	    break;
	}
    }

    count = matched.used;
    userList_clean(&matched);
    return count;
}

static int
trace_print_func(struct userNode *match, void *extra)
{
    struct discrim_and_source *das = extra;
    opserv_notice(das->source, "%s!%s@%s", match->nick, match->username, match->hostname);
    return 0;
}

static int
trace_count_func(struct userNode *match, void *extra)
{
    (void)match;(void)extra;
    return 0;
}

static int
trace_scan_func(struct userNode *match, void *extra)
{
    (void)extra;
    sockcheck_queue_address(htonl(match->ip));
    return 0;
}

static int
is_oper_victim(struct userNode *user, struct userNode *target, int match_opers)
{
    return !(IsService(target)
             || (!match_opers && IsOper(target))
             || (target->handle_info
                 && target->handle_info->opserv_level > user->handle_info->opserv_level));
}

static int
trace_gline_func(struct userNode *match, void *extra)
{
    struct discrim_and_source *das = extra;

    /* See the comment in trace_kill_func() below. */
    if(das->source == match) {
	return 0;
    }

    if (is_oper_victim(das->source, match, das->discrim->match_opers)) {
        opserv_block(match, das->source->handle_info->handle, das->discrim->reason);
    }

    return 0;
}

static int
trace_kill_func(struct userNode *match, void *extra)
{
    struct discrim_and_source *das = extra;

    /* If the person who issues a trace kill is among its victims,
       is_oper_victim will dereference a pointer to a user that doesn't
       exist anymore. We skip killing the issuing user to avoid this. */
    if(das->source == match) {
	return 0;
    }

    if (is_oper_victim(das->source, match, das->discrim->match_opers)) {
	char *reason;
        if (das->discrim->reason) {
            reason = das->discrim->reason;
        } else {
            reason = alloca(strlen(OSMSG_KILL_REQUESTED)+strlen(das->source->nick)+1);
            sprintf(reason, OSMSG_KILL_REQUESTED, das->source->nick);
        }
	DelUser(match, opserv, true, reason);
    }

    return 0;
}

static int
is_gagged(char *mask)
{
    struct gag_entry *gag;

    for (gag = gagList; gag; gag = gag->next) {
        if (match_ircglobs(gag->mask, mask)) return 1;
    }
    return 0;
}

static int
trace_gag_func(struct userNode *match, void *extra)
{
    struct discrim_and_source *das = extra;

    if (is_oper_victim(das->source, match, das->discrim->match_opers)) {
        char *reason, *mask;
        int masksize;
        if (das->discrim->reason) {
            reason = das->discrim->reason;
        } else {
            reason = alloca(strlen(OSMSG_GAG_REQUESTED)+strlen(das->source->nick)+1);
            sprintf(reason, OSMSG_GAG_REQUESTED, das->source->nick);
        }
	masksize = 5+strlen(match->hostname);
	mask = alloca(masksize);
        snprintf(mask, masksize, "*!*@%s", match->hostname);
	if (!is_gagged(mask)) {
            gag_create(mask, das->source->handle_info->handle, reason, 0);
        }
    }

    return 0;
}

static OPSERV_FUNC(cmd_trace)
{
    struct discrim_and_source das;
    discrim_search_func action;
    unsigned int matches;
    struct opserv_cmd *subcmd;

    OPSERV_MIN_PARMS(3, false);

    if (!strcasecmp(argv[1], "print")) {
	action = trace_print_func;
        subcmd = dict_find(opserv_func_dict, "trace print", NULL);
    } else if (!strcasecmp(argv[1], "count")) {
	action = trace_count_func;
        subcmd = dict_find(opserv_func_dict, "trace count", NULL);
    } else if (SOCKCHECK_USABLE && !strcasecmp(argv[1], "scan")) {
	action = trace_scan_func;
        subcmd = dict_find(opserv_func_dict, "trace scan", NULL);
    } else if (!strcasecmp(argv[1], "gline")) {
	action = trace_gline_func;
        subcmd = dict_find(opserv_func_dict, "trace gline", NULL);
    } else if (!strcasecmp(argv[1], "kill")) {
	action = trace_kill_func;
        subcmd = dict_find(opserv_func_dict, "trace kill", NULL);
    } else if (!strcasecmp(argv[1], "gag")) {
	action = trace_gag_func;
	subcmd = dict_find(opserv_func_dict, "trace gag", NULL);
    } else {
	opserv_notice(user, OSMSG_BAD_ACTION, argv[1]);
	return 0;
    }

    if (user->handle_info->opserv_level < subcmd->min_level) {
        opserv_notice(user, OSMSG_LEVEL_TOO_LOW);
        return 0;
    }

    das.discrim = opserv_discrim_create(user, argc-2, argv+2);
    if (!das.discrim) return 0;
    das.source = user;

    if (action == trace_print_func) {
	opserv_notice(user, MSG_SEARCH_RESULTS, "users");
    } else if (action == trace_count_func) {
	das.discrim->limit = INT_MAX;
    }

    matches = opserv_discrim_search(das.discrim, action, &das);

    if (matches) {
	opserv_notice(user, MSG_MATCH_COUNT, matches);
    } else {
	opserv_notice(user, MSG_NO_MATCHES);
    }

    free(das.discrim);
    return 1;
}

static OPSERV_FUNC(cmd_gsync)
{
    struct server *src;
    (void)channel;
    if (argc > 1) {
        src = GetServerH(argv[1]);
        if (!src) {
            opserv_notice(user, MSG_SERVER_UNKNOWN, argv[1]);
            return 0;
        }
    } else {
        src = self->uplink;
    }
    irc_stats(opserv, src, 'G');
    opserv_notice(user, OSMSG_GSYNC_RUNNING, src->name);
    return 1;
}

struct gline_extra {
    struct userNode *user;
    struct string_list *glines;
};

static void
gtrace_print_func(struct gline *gline, void *extra)
{
    struct gline_extra *xtra = extra;
    char *when_text = asctime(localtime(&gline->expires));
    when_text[strlen(when_text)-1] = 0; /* strip lame \n */
    opserv_notice(xtra->user, OSMSG_GTRACE_FORMAT, gline->target, gline->issuer, when_text, gline->reason);
}

static void
gtrace_count_func(struct gline *gline, void *extra)
{
    (void)gline; (void)extra;
}

static void
gtrace_ungline_func(struct gline *gline, void *extra)
{
    struct gline_extra *xtra = extra;
    string_list_append(xtra->glines, gline->target);
}

static OPSERV_FUNC(cmd_gtrace)
{
    struct gline_discrim *discrim;
    gline_search_func action;
    unsigned int matches;
    struct gline_extra extra;

    OPSERV_MIN_PARMS(3, false);
    extra.user = user;
    extra.glines = alloc_string_list(4);
    if (!strcasecmp(argv[1], "print")) {
        action = gtrace_print_func;
    } else if (!strcasecmp(argv[1], "count")) {
        action = gtrace_count_func;
    } else if (!strcasecmp(argv[1], "ungline")) {
        action = gtrace_ungline_func;
    } else {
        opserv_notice(user, OSMSG_BAD_ACTION, argv[1]);
        return 0;
    }

    discrim = gline_discrim_create(user, opserv, argc-2, argv+2);
    if (!discrim) return 0;

    if (action == gtrace_print_func) {
        opserv_notice(user, MSG_SEARCH_RESULTS, "glines");
    } else if (action == gtrace_count_func) {
        discrim->limit = INT_MAX;
    }

    matches = gline_discrim_search(discrim, action, &extra);

    if (action == gtrace_ungline_func) {
        unsigned int nn;
        for (nn=0; nn<extra.glines->used; nn++) {
            gline_remove(extra.glines->list[nn], 1);
        }
    }

    if (matches) {
        opserv_notice(user, MSG_MATCH_COUNT, matches);
    } else {
        opserv_notice(user, MSG_NO_MATCHES);
    }
    free(discrim);
    return 1;
}

static int
alert_check_user(const char *key, void *data, void *extra)
{
    struct opserv_user_alert *alert = data;
    struct userNode *user = extra;
    if (discrim_match(alert->discrim, user)) {
	switch (alert->reaction) {
	case REACT_NOTICE:
	    opserv_alert(OSMSG_ALERT_TRIGGERED, key, user->nick, alert->discrim->reason);
	    break;
        case REACT_GLINE:
            opserv_block(user, alert->owner, alert->discrim->reason);
            break;
	default:
	    opserv_alert(OSMSG_ALERT_TRIGGERED, key, user->nick, alert->discrim->reason);
            log(OS_LOG, LOG_ERROR, "Invalid reaction type %d for alert %s\n", alert->reaction, key);
	    break;
	}
    }
    return 0;
}

static int
opserv_alert_check(struct userNode *user)
{
    return dict_foreach(opserv_user_alerts, alert_check_user, user) ? 1 : 0;
}

static void
opserv_alert_check_nick(struct userNode *user, const char *new_nick)
{
    char old_nick[NICKLEN+1];
    strcpy(old_nick, user->nick);
    strcpy(user->nick, new_nick);
    dict_foreach(opserv_nick_based_alerts, alert_check_user, user);
    strcpy(user->nick, old_nick);
}

static void
opserv_staff_alert(struct userNode *user, struct handle_info *old_handle)
{
    const unsigned char *type;

    (void)old_handle;
    if (!opserv_conf.staff_auth_channel) return;
    if (!user->handle_info) return;
    else if (user->handle_info->opserv_level) type = "OPER";
    else if (IsHelper(user)) type = "HELPER";
    else return;

    if (user->ip) {
        send_channel_notice(opserv_conf.staff_auth_channel, opserv,
                            IDENT_FORMAT" authed to %s handle %s", IDENT_DATA(user), type, user->handle_info->handle);
    } else {
        send_channel_notice(opserv_conf.staff_auth_channel, opserv,
                            "%s [%s@%s] authed to %s handle %s", user->nick, user->username, user->hostname, type, user->handle_info->handle);
    }
}

static void
entry_search_helper_func(struct logEntry *match, void *data)
{
    char *service, *severity;
    static char timestamp[24];
    strftime(timestamp, sizeof(timestamp), "[%H:%M:%S %m/%d/%Y]", localtime(&match->time));

    switch (match->type) {
    case MAIN_LOG: service = "main"; break;
    case DEBUG_LOG: service = "debug"; break;
    case CS_LOG: service = "$C"; break;
    case NS_LOG: service = "$N"; break;
    case OS_LOG: service = "$O"; break;
    case PC_LOG: service = "proxycheck"; break;
    default: service = "unknown"; break;
    }

    switch (match->severity) {
    case LOG_INFO: severity = "(info) "; break;
    case LOG_WARNING: severity = "(warning) "; break;
    case LOG_ERROR: severity = "(error) "; break;
    case LOG_OTHER: default: severity = ""; break;
    }


    opserv_notice(data, "%s: %s %s%s", service, timestamp, severity, match->text);
}

static OPSERV_FUNC(cmd_log)
{
    struct logSearch *discrim;
    unsigned int matches;

    OPSERV_MIN_PARMS(2, false);

    discrim = log_discrim_create(opserv, user, argc, argv);
    if (!discrim) return 0;

    opserv_notice(user, MSG_SEARCH_RESULTS, "log entries");
    matches = log_entry_search(discrim, entry_search_helper_func, user);

    if (matches) {
	opserv_notice(user, MSG_MATCH_COUNT, matches);
    } else {
	opserv_notice(user, MSG_NO_MATCHES);
    }

    free(discrim);
    return 1;
}

static int
gag_helper_func(struct userNode *match, void *extra)
{
    (void)extra;

    /* TODO: may want to create a list of hosts to protect from glines, gags
       etc.. we could then test for such hosts here. */
    if(IsPrivileged(match) || IsLocal(match)) return 0;
    match->modes |= FLAGS_GAGGED;
    return 0;
}

static int
opserv_gag_check(struct userNode *user)
{
    struct gag_entry *gag;

    for (gag = gagList; gag; gag = gag->next) {
	if (user_matches_glob(user, gag->mask, 1)) {
	    gag_helper_func(user, NULL);
	    break;
	}
    }
    return 0;
}

static OPSERV_FUNC(cmd_timedgag)
{
    struct gag_entry *gag;
    unsigned int gagged;

    OPSERV_MIN_PARMS(4, false);
    unsplit_string(argv + 3, argc - 3);

    if (!is_ircmask(argv[1])) {
	opserv_notice(user, OSMSG_INVALID_IRCMASK, argv[1]);
        return 0;
    }

    for (gag = gagList; gag; gag = gag->next) {
	if (match_ircglobs(gag->mask, argv[1])) break;
    }

    if (gag) {
	opserv_notice(user, OSMSG_REDUNDANT_GAG, argv[1]);
	return 0;
    }

    gagged = gag_create(argv[1], user->handle_info->handle, argv[3], now+ParseInterval(argv[2]));

    if (gagged) {
	opserv_notice(user, OSMSG_GAG_APPLIED, argv[1], gagged);
    } else {
	opserv_notice(user, OSMSG_GAG_ADDED, argv[1]);
    }

    return 1;
}

static OPSERV_FUNC(cmd_gag)
{
    struct gag_entry *gag;
    unsigned int gagged;

    OPSERV_MIN_PARMS(3, false);
    unsplit_string(argv + 2, argc - 2);

    if (!is_ircmask(argv[1])) {
	opserv_notice(user, OSMSG_INVALID_IRCMASK, argv[1]);
        return 0;
    }

    for (gag = gagList; gag; gag = gag->next) {
	if (match_ircglobs(gag->mask, argv[1])) break;
    }

    if (gag) {
	opserv_notice(user, OSMSG_REDUNDANT_GAG, argv[1]);
	return 0;
    }

    gagged = gag_create(argv[1], user->handle_info->handle, argv[2], 0);

    if (gagged) {
	opserv_notice(user, OSMSG_GAG_APPLIED, argv[1], gagged);
    } else {
	opserv_notice(user, OSMSG_GAG_ADDED, argv[1]);
    }

    return 1;
}

static int
ungag_helper_func(struct userNode *match, void *extra)
{
    (void)extra;
    match->modes &= ~FLAGS_GAGGED;
    return 0;
}

static OPSERV_FUNC(cmd_ungag)
{
    struct gag_entry *gag;
    unsigned int ungagged;

    OPSERV_MIN_PARMS(2, false);

    for (gag = gagList; gag; gag = gag->next) {
	if (!strcmp(gag->mask, argv[1])) break;
    }

    if (!gag) {
	opserv_notice(user, OSMSG_GAG_NOT_FOUND, argv[1]);
	return 0;
    }

    timeq_del(gag->expires, gag_expire, gag, 0);
    ungagged = gag_free(gag);

    if (ungagged) {
	opserv_notice(user, OSMSG_UNGAG_APPLIED, argv[1], ungagged);
    } else {
	opserv_notice(user, OSMSG_UNGAG_ADDED, argv[1]);
    }

    return 1;
}

static OPSERV_FUNC(cmd_alias)
{
    int action;

    OPSERV_MIN_PARMS(4, false);

    action = add_alias(argv[1], argv[2], argv[3]);
    switch (action) {
    case ALIAS_CREATED:
	opserv_notice(user, OSMSG_ALIAS_CREATED, argv[1], argv[2], argv[3]);
	break;
    case ALIAS_MODIFIED:
	opserv_notice(user, OSMSG_ALIAS_MODIFIED, argv[1], argv[2], argv[3]);
	break;
    case ALIAS_SERVICE_NOT_FOUND:
	opserv_notice(user, OSMSG_NO_SUCH_SERVICE, argv[1]);
	break;
    case ALIAS_COMMAND_NOT_FOUND:
	opserv_notice(user, OSMSG_NO_SUCH_COMMAND, argv[1], argv[3]);
	break;
    case ALIAS_FOUND_COMMAND:
	opserv_notice(user, OSMSG_WONT_ALIAS_CMD, argv[1], argv[2]);
	break;
    case ALIAS_SELF:
	opserv_notice(user, OSMSG_SELF_ALIAS);
	break;
    default:
	opserv_notice(user, OSMSG_ALIAS_CREATE_FAIL, argv[1], argv[2]);
    }

    return 1;
}

static OPSERV_FUNC(cmd_unalias)
{
    int action;

    OPSERV_MIN_PARMS(3, false);

    action = del_alias(argv[1], argv[2]);
    switch (action) {
    case ALIAS_DELETED:
	opserv_notice(user, OSMSG_ALIAS_DELETED, argv[1], argv[2]);
	break;
    case ALIAS_SERVICE_NOT_FOUND:
	opserv_notice(user, OSMSG_NO_SUCH_SERVICE, argv[1]);
	break;
    case ALIAS_NOT_FOUND:
	opserv_notice(user, OSMSG_NO_SUCH_ALIAS, argv[1], argv[2]);
	break;
    case ALIAS_FOUND_COMMAND:
	opserv_notice(user, OSMSG_WONT_ALIAS_CMD, argv[1], argv[3]);
	break;
    default:
	opserv_notice(user, OSMSG_ALIAS_DELETED, argv[1], argv[3]);
    }

    return 1;
}

static int
command_aliases_helper(const char *key, void *data, void *extra)
{
    char *aliases = cNode_aliases(data), *c;
    (void)data;

    if (!aliases) {
	opserv_notice(extra, OSMSG_COMMAND_ALIASES, key, "None.");
	return 0;
    }

    for (c = aliases; *c; c++) *c = toupper(*c);

    opserv_notice(extra, OSMSG_COMMAND_ALIASES, key, aliases);
    free(aliases);

    return 0;
}

static int
service_commands_helper(const char *key, void *data, void *extra)
{
    struct serviceNode *sNode = data;

    opserv_notice(extra, OSMSG_SERVICE_COMMANDS, key);
    if (!dict_size(sNode->commands)) {
	opserv_notice(extra, "$bNone.$b");
    } else {
	dict_foreach(sNode->commands, command_aliases_helper, extra);
    }

    return 0;
}

static OPSERV_FUNC(cmd_aliases)
{
    extern dict_t services;
    (void)channel;

    if (argc > 1) {
	struct serviceNode *sNode = dict_find(services, argv[1], NULL);
	if (!sNode) {
	    opserv_notice(user, OSMSG_NO_SUCH_SERVICE, argv[1]);
	    return 0;
	}

	service_commands_helper(sNode->service, sNode, user);
	return 1;
    }

    dict_foreach(services, service_commands_helper, user);
    return 1;
}

static OPSERV_FUNC(cmd_addalert)
{
    const char *name;
    opserv_alert_reaction reaction;
    struct opserv_cmd *subcmd;

    OPSERV_MIN_PARMS(4, 0);
    name = argv[1];
    if (!strcasecmp(argv[2], "notice")) {
	reaction = REACT_NOTICE;
        subcmd = dict_find(opserv_func_dict, "addalert notice", NULL);
    } else if (!strcasecmp(argv[2], "gline")) {
        reaction = REACT_GLINE;
        subcmd = dict_find(opserv_func_dict, "addalert gline", NULL);
    } else {
	opserv_notice(user, OSMSG_UNKNOWN_REACTION, argv[2]);
	return 0;
    }
    if (user->handle_info->opserv_level < subcmd->min_level) {
        opserv_notice(user, OSMSG_LEVEL_TOO_LOW);
        return 0;
    }
    unsplit_string(argv + 3, argc - 3);
    if (opserv_add_user_alert(user, name, reaction, argv[3])) {
	opserv_notice(user, OSMSG_ADDED_ALERT, name);
    }
    return 1;
}

static OPSERV_FUNC(cmd_delalert)
{
    unsigned int i;
    OPSERV_MIN_PARMS(2, 0);
    for (i=1; i<argc; i++) {
	if (dict_remove(opserv_user_alerts, argv[i])) {
            dict_remove(opserv_nick_based_alerts, argv[i]);
	    opserv_notice(user, OSMSG_REMOVED_ALERT, argv[i]);
	} else {
	    opserv_notice(user, OSMSG_NO_SUCH_ALERT, argv[i]);
	}
    }
    return 1;
}

static OPSERV_FUNC(cmd_timecmd)
{
    struct timeval start, stop;
    OPSERV_MIN_PARMS(2, 0);
    unsplit_string(argv+1, argc-1);
    gettimeofday(&start, NULL);
    opserv_chanmsg(user, channel, argv[1]);
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
	stop.tv_sec -= 1;
	stop.tv_usec += 1000000;
    }
    opserv_notice(user, OSMSG_COMMAND_TIME, argv[1], stop.tv_sec, stop.tv_usec);
    return 1;
}

static OPSERV_FUNC(cmd_modlog)
{
    OPSERV_MIN_PARMS(3, 0);
    if (log_modify_logging(argv[1], argv[2])) {
        opserv_notice(user, OSMSG_MODLOG_PASSED);
        return 1;
    } else {
        opserv_notice(user, OSMSG_MODLOG_FAILED);
        return 0;
    }
}

static OPSERV_FUNC(cmd_loginfo)
{
    char buffer[MAXLEN], *sevname;
    int mask, buffer_pos = 0, sev_len;
    enum log_severity sev;
    OPSERV_MIN_PARMS(2, 0);
    mask = log_get_mask(argv[1]);
    for (buffer_pos = 0, sev = LOG_OTHER;
         sev < LOG_NUM_SEVERITIES;
         sev++) {
        if (!(mask & (1 << sev))) continue;
        switch (sev) {
        case LOG_OTHER: sevname = "other"; break;
        case LOG_INFO: sevname = "info"; break;
        case LOG_WARNING: sevname = "warning"; break;
        case LOG_ERROR: sevname = "error"; break;
        default: sevname = "(unknown)"; break;
        }
        sev_len = strlen(sevname);
        memcpy(buffer+buffer_pos, sevname, sev_len);
        buffer_pos += sev_len;
        buffer[buffer_pos++] = ' ';
    }
    buffer[buffer_pos] = 0;
    opserv_notice(user, OSMSG_LOG_SEVERITIES, argv[1], buffer);
    return 1;
}

static OPSERV_FUNC(cmd_sanity)
{
    char *res;
    (void)channel; (void)argc; (void)argv;
    if ((res = dict_sanity_check(clients))) {
        opserv_notice(user, "Clients insanity: %s", res);
        free(res);
    } else {
        opserv_notice(user, "Clients sanity: check");
    }
    if ((res = dict_sanity_check(channels))) {
        opserv_notice(user, "Channels insanity: %s", res);
        free(res);
    } else {
        opserv_notice(user, "Channels sanity: check");
    }
    return 1;
}

static int
set_join_policer(const char *key, void *data, void *extra)
{
    struct chanNode *cNode = data;
    (void)key;

    if (!cNode->join_policer) {
	cNode->join_policer = policer_new(extra, now);
    }

    return 0;
}

static void
opserv_conf_read(void)
{
    struct record_data *rd;
    dict_t conf_node, child;
    const char *str;
    struct policer_params *pp;

    rd = conf_get_node(OPSERV_CONF_NAME);
    if (!rd || rd->type != RECDB_OBJECT) {
	log(OS_LOG, LOG_ERROR, "config node `%s' is missing or has wrong type\n",
	    OPSERV_CONF_NAME);
	return;
    }
    conf_node = rd->d.object;
    str = database_get_data(conf_node, KEY_DB_BACKUP_FREQ, RECDB_QSTRING);
    opserv_conf.db_backup_frequency = str ? ParseInterval(str) : 7200;
    str = database_get_data(conf_node, KEY_DEBUG_CHANNEL, RECDB_QSTRING);
    if (str) {
	opserv_conf.debug_channel = AddChannel(str, now, "+tinms", NULL, 0, opserv->numeric, NULL);
    } else {
	opserv_conf.debug_channel = NULL;
    }
    str = database_get_data(conf_node, KEY_ALERT_CHANNEL, RECDB_QSTRING);
    if (str) {
	opserv_conf.alert_channel = AddChannel(str, now, "+tns", NULL, 0, opserv->numeric, NULL);
    } else {
	opserv_conf.alert_channel = NULL;
    }
    str = database_get_data(conf_node, KEY_STAFF_AUTH_CHANNEL, RECDB_QSTRING);
    if (str) {
        opserv_conf.staff_auth_channel = AddChannel(str, now, "+timns", NULL, 0, opserv->numeric, NULL);
    } else {
        opserv_conf.staff_auth_channel = NULL;
    }
    str = database_get_data(conf_node, KEY_UNTRUSTED_MAX, RECDB_QSTRING);
    opserv_conf.untrusted_max = str ? strtoul(str, NULL, 0) : 5;
    str = database_get_data(conf_node, KEY_PURGE_LOCK_DELAY, RECDB_QSTRING);
    opserv_conf.purge_lock_delay = str ? strtoul(str, NULL, 0) : 60;
    str = database_get_data(conf_node, KEY_NICK, RECDB_QSTRING);
    if (str) NickChange(opserv, str);
    str = database_get_data(conf_node, KEY_TRIGGER, RECDB_QSTRING);
    opserv_conf.trigger_char = str ? str[0] : '?';

    opserv_conf.join_policer_params = policer_params_new();
    if ((child = database_get_data(conf_node, KEY_JOIN_POLICER, RECDB_OBJECT))) {
	dict_foreach(child, set_policer_param, opserv_conf.join_policer_params);
    } else {
	policer_params_set(opserv_conf.join_policer_params, "size", "20");
	policer_params_set(opserv_conf.join_policer_params, "drain-rate", "1");
    }

    dict_foreach(channels, set_join_policer, opserv_conf.join_policer_params);

    pp = policer_params_new();
    if ((child = database_get_data(conf_node, KEY_NEW_USER_POLICER, RECDB_OBJECT))) {
	dict_foreach(child, set_policer_param, pp);
    } else {
	policer_params_set(pp, "size", "200");
	policer_params_set(pp, "drain-rate", "3");
    }
    opserv_conf.new_user_policer = policer_new(pp, now);
}

static void
opserv_db_cleanup(void)
{
    close_helpfile(opserv_helpfile);
    dict_delete(opserv_chan_warn);
    dict_delete(opserv_func_dict);
    dict_delete(opserv_reserved_nick_dict);
    free_string_list(opserv_bad_words);
    free_string_list(opserv_secret_words);
    dict_delete(opserv_trusted_hosts);
    unreg_del_user_func(opserv_user_cleanup);
    dict_delete(opserv_hostinfo_dict);
    dict_delete(opserv_user_alerts);
    dict_delete(opserv_nick_based_alerts);
}

void
init_opserv(const char *nick)
{
    opserv = AddService(nick, "Oper Services");
    opserv_conf_read();
    conf_register_reload(opserv_conf_read);

    reg_privmsg_func(opserv->num_local, opserv_privmsg);
    reg_chanmsg_func(opserv_conf.trigger_char, opserv, opserv_chanmsg);

    opserv_func_dict = dict_new();
    opserv_define_func("ACCESS", cmd_access, 0);
    opserv_define_func("ADDALERT", cmd_addalert, 800);
    opserv_define_func("ADDALERT NOTICE", NULL, 0);
    opserv_define_func("ADDALERT GLINE", NULL, 900);
    opserv_define_func("ADDBAD", cmd_addbad, 800);
    opserv_define_func("ADDSECRET", cmd_addsecret, 800);
    opserv_define_func("ADDTRUST", cmd_addtrust, 800);
    opserv_define_func("ALIAS", cmd_alias, 800);
    opserv_define_func("ALIASES", cmd_aliases, 0);
    opserv_define_func("BAN", cmd_ban, 100);
    opserv_define_func("BLOCK", cmd_block, 100);
    opserv_define_func("CHANINFO", cmd_chaninfo, 0);
    opserv_define_func("CLEARBANS", cmd_clearbans, 300);
    opserv_define_func("CLEARMODES", cmd_clearmodes, 400);
    opserv_define_func("CHLEVEL", cmd_chlevel, 999);
    opserv_define_func("CLONE", cmd_clone, 999);
    opserv_define_func("COLLIDE", cmd_collide, 800);
    opserv_define_func("DELALERT", cmd_delalert, 800);
    opserv_define_func("DELBAD", cmd_delbad, 800);
    opserv_define_func("DELSECRET", cmd_delsecret, 800);
    opserv_define_func("DELTRUST", cmd_deltrust, 800);
    opserv_define_func("DEOP", cmd_deop, 100);
    opserv_define_func("DEOPALL", cmd_deopall, 400);
    opserv_define_func("DIE", cmd_die, 900);
    opserv_define_func("DISABLECOM", cmd_disablecom, 950);
    opserv_define_func("DUMP", cmd_dump, 999);
    opserv_define_func("ENABLECOM", cmd_enablecom, 950);
    opserv_define_func("GAG", cmd_gag, 600);
    opserv_define_func("GLINE", cmd_gline, 600);
    opserv_define_func("HELP", cmd_help, 0);
    opserv_define_func("INVITEME", cmd_inviteme, 100);
    opserv_define_func("ISON", cmd_ison, 0);
    opserv_define_func("JOIN", cmd_join, 601);
    opserv_define_func("JUMP", cmd_jump, 900);
    opserv_define_func("JUPE", cmd_jupe, 900);
    opserv_define_func("KICK", cmd_kick, 100);
    opserv_define_func("KICKALL", cmd_kickall, 400);
    opserv_define_func("KICKBAN", cmd_kickban, 100);
    opserv_define_func("KICKBANALL", cmd_kickbanall, 450);
    opserv_define_func("LOG", cmd_log, 900);
    opserv_define_func("LOGINFO", cmd_loginfo, 0);
    opserv_define_func("MODE", cmd_mode, 100);
    opserv_define_func("MODERATE", cmd_moderate, 300);
    opserv_define_func("MODLOG", cmd_modlog, 900);
    opserv_define_func("NICKBAN", cmd_nickban, 100);
    opserv_define_func("OP", cmd_op, 100);
    opserv_define_func("OPALL", cmd_opall, 400);
    opserv_define_func("PART", cmd_part, 601);
    opserv_define_func("QUERY", cmd_query, 0);
    opserv_define_func("RAW", cmd_raw, 999);
    opserv_define_func("READHELP", cmd_readhelp, 900);
    opserv_define_func("RECONNECT", cmd_reconnect, 900);
    opserv_define_func("REFRESHG", cmd_refreshg, 600);
    opserv_define_func("REHASH", cmd_rehash, 900);
    opserv_define_func("REOPEN", cmd_reopen, 900);
    opserv_define_func("RESERVE", cmd_reserve, 800);
    opserv_define_func("RESTART", cmd_restart, 900);
    opserv_define_func("SANITY", cmd_sanity, 100);
    opserv_define_func("SET", cmd_set, 900);
    opserv_define_func("SHOWDISCOM", cmd_showdiscom, 600);
    opserv_define_func("SHOWLEVEL", cmd_cmdlevel, 0);
    opserv_define_func("STATS", cmd_stats, 101);
    opserv_define_func("TIMEDGAG", cmd_timedgag, 600);
    opserv_define_func("TRACE", cmd_trace, 100);
    opserv_define_func("TRACE PRINT", NULL, 0);
    opserv_define_func("TRACE COUNT", NULL, 0);
    opserv_define_func("TRACE SCAN", NULL, 650);
    opserv_define_func("TRACE GLINE", NULL, 600);
    opserv_define_func("TRACE GAG", NULL, 600);
    opserv_define_func("TRACE KILL", NULL, 600);
    opserv_define_func("GSYNC", cmd_gsync, 600);
    opserv_define_func("GTRACE", cmd_gtrace, 100);
    opserv_define_func("TIMECMD", cmd_timecmd, 1);
    opserv_define_func("UNALIAS", cmd_unalias, 800);
    opserv_define_func("UNBAN", cmd_unban, 100);
    opserv_define_func("UNGAG", cmd_ungag, 600);
    opserv_define_func("UNGLINE", cmd_ungline, 600);
    opserv_define_func("UNJUPE", cmd_unjupe, 900);
    opserv_define_func("UNMODERATE", cmd_unmoderate, 300);
    opserv_define_func("UNRESERVE", cmd_unreserve, 800);
    opserv_define_func("UNWARN", cmd_unwarn, 800);
    opserv_define_func("VERSION", cmd_version, 0);
    opserv_define_func("WARN", cmd_warn, 800);
    opserv_define_func("WARNLIST", cmd_warnlist, 0);
    opserv_define_func("WHOIS", cmd_whois, 0);
    opserv_define_func("WRITE", cmd_write, 800);
    opserv_define_func("WRITEALL", cmd_writeall, 800);

    if (SOCKCHECK_USABLE) {
        reg_new_user_func(opserv_proxy_check);
        opserv_define_func("CLEARHOST", cmd_clearhost, 650);
        opserv_define_func("DEFPROXY", cmd_defproxy, 999);
        opserv_define_func("HOSTSCAN", cmd_hostscan, 650);
    } else {
        opserv_define_func("CLEARHOST", opserv_missing_support, 650);
        opserv_define_func("DEFPROXY", opserv_missing_support, 999);
        opserv_define_func("HOSTSCAN", opserv_missing_support, 650);
    }

    opserv_reserved_nick_dict = dict_new();
    opserv_hostinfo_dict = dict_new();
    dict_set_free_keys(opserv_hostinfo_dict, free);
    dict_set_free_data(opserv_hostinfo_dict, free);

    reg_new_user_func(opserv_trusted_host_check);
    reg_del_user_func(opserv_user_cleanup);
    reg_new_channel_func(opserv_channel_check);
    reg_del_channel_func(opserv_del_channel);
    reg_join_func(opserv_join_check);
    reg_new_user_func(opserv_gag_check);
    reg_new_user_func(opserv_nick_flood_check);
    reg_new_user_func(opserv_alert_check);
    reg_auth_func(opserv_staff_alert);
    reg_nick_change_func(opserv_alert_check_nick);

    alias_register_service("OpServ", opserv_command_search, opserv_add_alias, opserv_del_alias);

    opserv_db_read();
    opserv_helpfile_read();

    if (opserv_conf.db_backup_frequency) {
	timeq_add(now + opserv_conf.db_backup_frequency, opserv_timed_write, NULL);
    }

    reg_exit_func(opserv_db_cleanup);
    reg_db_write_func(opserv_db_write);
}
