/* timeq.c - time-based event queue
 * Copyright 2000 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: timeq.c,v 1.7.2.5 2002/01/21 01:50:15 entrope Exp $
 */

#include "config.h"
#include <stdlib.h>

#include "common.h"
#if HAVE_PTHREAD_H && SOCKCHECK_USABLE /* we could have multiple threads setting glines */
#include <pthread.h>
/* Sigh.  Want recursive mutexes in standard!
 * (The timeq mutex needs to be recursive since callers frequently do timeq_add from
 * within timeq_run() callbacks.) */
static pthread_mutex_t timeq_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t timeq_condvar = PTHREAD_COND_INITIALIZER;
static pthread_t timeq_locker;
static int timeq_mutex_count;

static void
timeq_mutex_lock(void) {
    pthread_mutex_lock(&timeq_mutex);
    if ((timeq_mutex_count > 0) && (timeq_locker != pthread_self())) {
        pthread_cond_wait(&timeq_condvar, &timeq_mutex);
    }
    timeq_locker = pthread_self();
    timeq_mutex_count++;
    pthread_mutex_unlock(&timeq_mutex);
}

static void
timeq_mutex_unlock(void) {
    pthread_mutex_lock(&timeq_mutex);
    timeq_mutex_count--;
    if (!timeq_mutex_count) pthread_cond_signal(&timeq_condvar);
    pthread_mutex_unlock(&timeq_mutex);
}

#else
#define timeq_mutex_lock() /*noop*/
#define timeq_mutex_unlock() /*noop*/
#endif

#include "heap.h"
#include "timeq.h"

heap_t timeq;

struct timeq_entry {
    timeq_func func;
    void *data;
};

void
timeq_init(void)
{
    timeq = heap_new(int_comparator);
#if HAVE_PTHREAD_H && SOCKCHECK_USABLE
    pthread_cond_signal(&timeq_condvar);
#endif
}

time_t
timeq_next(void)
{
    void *time;
    timeq_mutex_lock();
    heap_peek(timeq, &time, 0);
    timeq_mutex_unlock();
    return TIME_T_CAST(time);
}

void
timeq_add(time_t when, timeq_func func, void *data)
{
    struct timeq_entry *ent;
    void *w;
    ent = malloc(sizeof(struct timeq_entry));
    ent->func = func;
    ent->data = data;
    w = (void*)TIME_T_CAST(when);
    timeq_mutex_lock();
    heap_insert(timeq, w, ent);
    timeq_mutex_unlock();
}

struct timeq_extra {
    time_t when;
    timeq_func func;
    void *data;
    int mask;
};

static int
timeq_del_matching(void *key, void *data, void *extra)
{
    struct timeq_entry *a = data;
    struct timeq_extra *b = extra;
    if (((b->mask & TIMEQ_IGNORE_WHEN) || (TIME_T_CAST(key) == b->when))
	&& ((b->mask & TIMEQ_IGNORE_FUNC) || (a->func == b->func))
	&& ((b->mask & TIMEQ_IGNORE_DATA) || (a->data == b->data))) {
        free(data);
        return 1;
    } else {
        return 0;
    }
}

void
timeq_del(time_t when, timeq_func func, void *data, int mask)
{
    struct timeq_extra extra;
    extra.when = when;
    extra.func = func;
    extra.data = data;
    extra.mask = mask;
    timeq_mutex_lock();
    heap_remove_pred(timeq, timeq_del_matching, &extra);
    timeq_mutex_unlock();
}

unsigned int
timeq_size(void)
{
    return heap_size(timeq);
}

void
timeq_run(time_t now)
{
    void *k, *d;
    struct timeq_entry *ent;
    timeq_mutex_lock();
    while (heap_size(timeq) > 0) {
	heap_peek(timeq, &k, &d);
	if (TIME_T_CAST(k) > now) break;
	ent = d;
	ent->func(ent->data);
	heap_pop(timeq);
        free(ent);
    }
    timeq_mutex_unlock();
}
